import os
import textwrap
import sys
import stat
import copy

import esm_environment

from esm_parser import find_variable, user_error, user_note
from . import helpers
from . import dataprocess
from . import prepare
from .slurm import Slurm
from .pbs import Pbs

known_batch_systems = ["slurm", "pbs"]
reserved_jobtypes = ["prepcompute", "compute", "prepare", "tidy", "inspect"]


class UnknownBatchSystemError(Exception):
    """Raise this exception when an unknown batch system is encountered"""


class batch_system:

    # all wrappers to slurm, pbs and co as esm_runscript
    # should be written independent of actual batch system
    def __init__(self, config, name):
        self.name = name
        if name == "slurm":
            self.bs = Slurm(config)
        elif name == "pbs":
            self.bs = Pbs(config)
        else:
            raise UnknownBatchSystemError(name)

    def check_if_submitted(self):
        return self.bs.check_if_submitted()

    def get_jobid(self):
        return self.bs.get_jobid()

    def get_job_state(self, jobid):
        return self.bs.get_job_state(jobid)

    def job_is_still_running(self, jobid):
        return self.bs.job_is_still_running(jobid)

    def add_pre_launcher_lines(self, config, cluster, runfile):
        return self.bs.add_pre_launcher_lines(config, cluster, runfile)

    # TODO: remove it once it's not needed anymore (substituted by packjob)
    def write_het_par_wrappers(self, config):
        return self.bs.write_het_par_wrappers(config)

    def het_par_headers(self, config, cluster, all_values_flat):
        return self.bs.het_par_headers(config, cluster, all_values_flat)

    def prepare_launcher(self, config, cluster):
        self.bs.prepare_launcher(config, cluster)
        return config

    @staticmethod
    def get_run_filename(config, cluster):
        folder = config["general"]["thisrun_scripts_dir"]
        expid = config["general"]["expid"]
        startdate = config["general"]["current_date"]
        enddate = config["general"]["end_date"]
        run_filename = (
            f"{folder}/{expid}_{cluster}" f"_{config['general']['run_datestamp']}.run"
        )
        return run_filename

    @staticmethod
    def get_shell_header(config, cluster):
        header = []
        coupling_dir = os.path.dirname(os.path.realpath(__file__))

        this_batch_system = config["computer"]
        if "sh_interpreter" in this_batch_system:
            header.append("#!" + this_batch_system["sh_interpreter"])

        header.append(". " + coupling_dir + "/coupling/coupling_general.functions")
        return header

    @staticmethod
    def get_batch_header(config, cluster):
        header = []
        this_batch_system = config["computer"]
        if "sh_interpreter" in this_batch_system:
            header.append("#!" + this_batch_system["sh_interpreter"])

        tasks = config["general"]["resubmit_tasks"]
        nodes = config["general"]["resubmit_nodes"]

        if config["computer"].get("heterogeneous_parallelization", False):
            tasks_nodes_flag = "nodes_flag"
        elif config["computer"]["batch_system"] in ["pbs"]:
            tasks_nodes_flag = "nodes_flag"
        else:
            tasks_nodes_flag = "tasks_flag"

        if cluster == "compute":
            partition = config["computer"]["partitions"]["compute"]["name"]
        else:
            partition = config["computer"]["partitions"]["pp"]["name"]

        replacement_tags = [
            ("@tasks@", tasks),
            ("@nodes@", nodes),
            ("@partition@", partition),
            ("@jobtype@", cluster),
        ]

        all_flags = [
            "partition_flag",
            "time_flag",
            tasks_nodes_flag,
            "output_flags",
            "name_flag",
        ]
        conditional_flags = [
            "accounting_flag",
            "notification_flag",
            "hyperthreading_flag",
            "additional_flags",
            "overcommit_flag",
        ]
        # ??? Do we need the exclusive flag?
        if config["general"]["jobtype"] in ["prepcompute", "tidy"]:
            conditional_flags.append("exclusive_flag")
        for flag in conditional_flags:
            if flag in this_batch_system:
                values_in_flag = []
                flag_value = this_batch_system[flag]

                if isinstance(flag_value, str):
                    values_in_flag.append(flag_value)
                elif isinstance(flag_value, list):
                    values_in_flag.extend(flag_value)

                # checking whether we have any empty values
                any_empty_values = any([item.strip() == "" for item in values_in_flag])
                if not any_empty_values:
                    all_flags.append(flag)

        # some items in `all_values` list might be lists, so flatten it
        all_values = [this_batch_system.get(flag) for flag in all_flags]
        all_values_flat = []
        for value in all_values:
            if isinstance(value, str):
                all_values_flat.append(value)
            elif isinstance(value, list):
                all_values_flat.extend(value)

        # Call the ``het_par_headers`` method to calculate the heterogeneous
        # parallelization flags, if necessary
        all_values_flat = config["general"]["batch"].het_par_headers(
            config, cluster, all_values_flat
        )

        # loop over all batch flag values and replace the tags
        for value in all_values_flat:
            for (tag, repl) in replacement_tags:
                value = value.replace(tag, str(repl))
            if this_batch_system.get("header_start") is not None:
                header.append(this_batch_system["header_start"] + " " + value)

        return header

    @staticmethod
    def calculate_requirements(config, cluster=None):
        # get number of tasks for the whole job to be submitted,
        # as well as number of start process and end process for each
        # component (in case a hostfile needs to be written)

        tasks = 0
        nodes = 0
        start_proc = 0
        end_proc = 0
        start_core = 0
        end_core = 0

        # if not explicitly stated for which cluster we need the
        # requirements, calculate them for the job we are already in

        if not cluster:
            cluster = config["general"]["jobtype"]

        if cluster in reserved_jobtypes:
            for model in config["general"]["valid_model_names"]:
                omp_num_threads = int(config[model].get("omp_num_threads", 1))

                if "nproc" in config[model]:
                    print(f"nproc: {config[model]['nproc']}")

                    # kh 21.04.22 multi group support added, i.e. using (nproc * mpi_num_groups) MPI processes to start a program multiple times
                    # (used for FESOM-REcoM tracer loop parallelization (MPI based))
                    mpi_num_groups = config[model].get("mpi_num_groups", 1)

# kh 22.06.22 adjust total number of MPI processes via mpi_num_groups at lowest level (nproc)
                    config[model]["nproc"] *= mpi_num_groups
                    config[model]["tasks"] = config[model]["nproc"]

                    # cores_per_node = config['computer']['cores_per_node']

                elif "nproca" in config[model] and "nprocb" in config[model]:
                    config[model]["tasks"] = (
                        config[model]["nproca"] * config[model]["nprocb"]
                    )
                    # end_proc = start_proc + int(config[model]["nproca"])*int(config[model]["nprocb"]) - 1
                    # KH 30.04.20: nprocrad is replaced by more flexible
                    # partitioning using nprocar and nprocbr
                    remove_nprocr_options = ["remove_from_namelist", 0]
                    if (
                        config[model].get("nprocar", 0) not in remove_nprocr_options
                        and config[model].get("nprocbr", 0) not in remove_nprocr_options
                    ):
                        config[model]["tasks"] += (
                            config[model]["nprocar"] * config[model]["nprocbr"]
                        )

                else:
                    continue

                # seb-wahl: add support for ECHAM6's parallel I/O feature
                # namelist parctl in namelist.echam
                if "nprocio" in config[model]:
                    config[model]["tasks"] += config[model].get("nprocio",0)

                nproc = config[model]["tasks"]
                if cluster == "compute":
                    cores_per_node = config["computer"]["partitions"]["compute"][
                        "cores_per_node"
                    ]
                else:
                    cores_per_node = config["computer"]["partitions"]["pp"][
                        "cores_per_node"
                    ]
                nodes += int(nproc * omp_num_threads / cores_per_node) + (
                    (nproc * omp_num_threads) % cores_per_node > 0
                )

                config[model]["threads"] = config[model]["tasks"] * omp_num_threads
                tasks += config[model]["tasks"]
                print(f"tasks: {tasks}")
                # Use the number of tasks and threads to update end_proc/core
                end_proc = start_proc + config[model]["tasks"] - 1
                end_core = start_core + config[model]["threads"] - 1
                # Feed the config with the final start_proc/core and end_proc/core
                config[model]["end_proc"] = end_proc
                config[model]["end_core"] = end_core
                config[model]["start_proc"] = start_proc
                config[model]["start_core"] = start_core
                # Reset the start_proc/core variables for the next component
                start_proc = end_proc + 1
                start_core = end_core + 1

        else:
            # dataprocessing job with user definded name
            # number of tasks are actually already prepared in
            # workflow

            if (
                not cluster
                or not cluster in config["general"]["workflow"]["subjob_clusters"]
            ):
                print(f"Unknown or unset cluster: {cluster}.")
                sys.exit(-1)
            # user defined jobtype doing dataprocessing
            tasks = config["general"]["workflow"]["subjob_clusters"][cluster]["nproc"]
            cores_per_node = config["computer"]["partitions"]["pp"]["cores_per_node"]
            nodes = int(tasks / cores_per_node) + ((tasks % cores_per_node) > 0)

        config["general"]["resubmit_tasks"] = tasks
        print(f"resubmit tasks: {config['general']['resubmit_tasks']}")
        config["general"]["resubmit_nodes"] = nodes

        return config

    @staticmethod
    def get_environment(config, subjob):
        environment = []

        env = esm_environment.environment_infos("runtime", config)
        commands = env.commands
        if not subjob.replace("_general", "") in reserved_jobtypes:  # ??? fishy
            commands += dataprocess.subjob_environment(config, subjob)
        commands += [""]

        return commands

    @staticmethod
    def get_extra(config):
        extras = []
        if config["general"].get("unlimited_stack_size", True):
            extras.append("# Set stack size to unlimited")
            extras.append("ulimit -s unlimited")
        if config["general"].get("use_venv", False):
            extras.append("# Start everything in a venv")
            extras.append(
                "source "
                + config["general"]["experiment_dir"]
                + "/.venv_esmtools/bin/activate"
            )
        if config["general"].get("funny_comment", True):
            extras.append("# 3...2...1...Liftoff!")
        # Search for ``pre_run_commands``s in the components
        for component in config.keys():
            pre_run_commands = config[component].get("pre_run_commands")
            if isinstance(pre_run_commands, list):
                for pr_command in pre_run_commands:
                    if isinstance(pr_command, str):
                        extras.append(pr_command)
                    else:
                        user_error(
                            'Invalid type for "pre_run_commands"',
                            (
                                f'"{type(pr_command)}" type is not supported for '
                                + f'elements of the "pre_run_commands", defined in '
                                + f'"{component}". Please, define '
                                + '"pre_run_commands" as a "list" of "strings" or a "list".'
                            ),
                        )
            elif isinstance(pre_run_commands, str):
                extras.append(pre_run_commands)
            elif pre_run_commands == None:
                continue
            else:
                user_error(
                    'Invalid type for "pre_run_commands"',
                    (
                        f'"{type(pre_run_commands)}" type is not supported for '
                        + f'"pre_run_commands" defined in "{component}". Please, define '
                        + '"pre_run_commands" as a "string" or a "list" of "strings".'
                    ),
                )
        return extras

    @staticmethod
    def append_start_statement(config, subjob):
        line = helpers.assemble_log_message(
            config,
            [
                subjob.replace("_general", ""),
                config["general"]["run_number"],
                config["general"]["current_date"],
                config["general"]["jobid"],
                "- start",
            ],
            timestampStr_from_Unix=True,
        )
        startline = "echo " + line + " >> " + config["general"]["experiment_log_file"]
        return startline

    @staticmethod
    def append_done_statement(config, subjob):
        line = helpers.assemble_log_message(
            config,
            [
                subjob.replace("_general", ""),
                config["general"]["run_number"],
                config["general"]["current_date"],
                config["general"]["jobid"],
                "- done",
            ],
            timestampStr_from_Unix=True,
        )
        doneline = "echo " + line + " >> " + config["general"]["experiment_log_file"]
        return doneline

    @staticmethod
    def get_run_commands(config, subjob, batch_or_shell):  # here or in compute.py?

        commands = []
        if subjob.startswith("compute"):
            if config["general"].get("submit_to_batch_system", True):
                batch_system = config["computer"]
                if "execution_command" in batch_system:
                    commands.append(
                        "time "
                        + batch_system["execution_command"]
                        + f" 2>&1{config['computer'].get('write_execution_log', '')} &"
                    )
                    if config["general"].get("multi_srun"):
                        return self.bs.get_run_commands_multisrun(config, commands)
            else:
                for model in config:
                    if model == "computer":
                        continue
                    if "execution_command" in config[model]:
                        commands.append(
                            "time ./"
                            + config[model]["execution_command"]
                            + f" 2>&1{config['computer'].get('write_execution_log', '')} &"
                        )
        else:
            subjob_tasks = dataprocess.subjob_tasks(config, subjob, batch_or_shell)
            for task in subjob_tasks:
                commands.append(task)

        return commands

    @staticmethod
    def get_submit_command(config, batch_or_shell, runfilename):
        # in case of slurm e.g. returns:
        # cd SCRIPTDIR; sbatch runfile
        # in case if shell:
        # cd SCRIPTDIR; ./runfile

        commands = []
        if batch_or_shell == "batch":
            call = config["computer"]["submit"] + " "
        else:
            call = "./"

        batch_system = config["computer"]

        if "submit" in batch_system:
            commands.append(
                "cd "
                + config["general"]["thisrun_scripts_dir"]
                + "; "
                + call
                + os.path.basename(runfilename)
            )

        return commands

    @staticmethod
    def write_simple_runscript(config, cluster, batch_or_shell="batch"):

        # if no cluster is specified, work on the one we are in
        # if not cluster:
        #    cluster = config["general"]["jobtype"]

        clusterconf = None
        if "workflow" in config["general"]:
            if "subjob_clusters" in config["general"]["workflow"]:
                if cluster in config["general"]["workflow"]["subjob_clusters"]:
                    clusterconf = config["general"]["workflow"]["subjob_clusters"][
                        cluster
                    ]

        if not clusterconf:
            print(f"No config found for cluster {cluster}.")
            sys.exit(-1)

        self = config["general"]["batch"]
        runfilename = batch_system.get_run_filename(config, cluster)

        if config["general"]["verbose"]:
            print("still alive")
            print("jobtype: ", config["general"]["jobtype"])
            print("writing run file for:", cluster)

        with open(runfilename, "w") as runfile:

            # batch header (if any)
            if batch_or_shell == "batch":

                config = batch_system.calculate_requirements(config, cluster)
                # TODO: remove it once it's not needed anymore (substituted by packjob)
                if (
                    cluster in reserved_jobtypes
                    and config["computer"].get("taskset", False)
                ):
                    config = config["general"]["batch"].write_het_par_wrappers(config)
                # Prepare launcher
                config = config["general"]["batch"].prepare_launcher(config, cluster)
                # Initiate the header
                header = batch_system.get_batch_header(config, cluster)

                for line in header:
                    runfile.write(line + "\n")
                runfile.write("\n")

            else:
                header = batch_system.get_shell_header(config, cluster)
                for line in header:
                    runfile.write(line + "\n")
                runfile.write("\n")

            if clusterconf:
                for subjob in clusterconf["subjobs"]:

                    # environment for each subjob of a cluster
                    environment = batch_system.get_environment(config, subjob)
                    batch_system.write_env(config, environment, runfilename)
                    for line in environment:
                        runfile.write(line + "\n")

                    # extra entries for each subjob
                    extra = batch_system.get_extra(config)
                    for line in extra:
                        runfile.write(line + "\n")

                    # Add actual commands
                    commands = batch_system.get_run_commands(
                        config, subjob, batch_or_shell
                    )
                    # commands = clusterconf.get("data_task_list", [])
                    runfile.write("\n")
                    runfile.write(self.append_start_statement(config, subjob) + "\n")
                    runfile.write("\n")
                    runfile.write("cd " + config["general"]["thisrun_work_dir"] + "\n")
                    if cluster in reserved_jobtypes:
                        config["general"]["batch"].add_pre_launcher_lines(
                            config, cluster, runfile
                        )

                    for line in commands:
                        runfile.write(line + "\n")

            # elif multisrun_stuff: # pauls stuff maybe here? or matching to clusterconf possible?
            #    dummy = 0
            else:  # "normal" case
                dummy = 0

            if submits_another_job(config, cluster):  # and batch_or_shell == "batch":
                # -j ? is that used somewhere? I don't think so, replaced by workflow
                #   " -j "+ config["general"]["jobtype"]

                observe_call = (
                    "esm_runscripts "
                    + config["general"]["scriptname"]
                    + " -e "
                    + config["general"]["expid"]
                    + " -t observe_"
                    + cluster
                    + " -p ${process}"
                    + " -s "
                    + config["general"]["current_date"].format(
                        form=9, givenph=False, givenpm=False, givenps=False
                    )
                    + " -r "
                    + str(config["general"]["run_number"])
                    + " -v "
                    + " --last-jobtype "
                    + config["general"]["jobtype"]
                )

                if "--open-run" in config["general"]["original_command"] or not config[
                    "general"
                ].get("use_venv"):
                    observe_call += " --open-run"
                elif "--contained-run" in config["general"][
                    "original_command"
                ] or config["general"].get("use_venv"):
                    observe_call += " --contained-run"
                else:
                    print("ERROR -- Not sure if you were in a contained or open run!")
                    print(
                        "ERROR -- See write_simple_runscript for the code causing this."
                    )
                    sys.exit(1)

                if "modify_config_file_abspath" in config["general"]:
                    if config["general"]["modify_config_file_abspath"]:
                        observe_call += (
                            " -m " + config["general"]["modify_config_file_abspath"]
                        )

                subjobs_to_launch = config["general"]["workflow"]["subjob_clusters"][
                    cluster
                ]["next_submit"]

                runfile.write("\n")
                runfile.write("# Call to esm_runscript to start subjobs:\n")
                runfile.write("# " + str(subjobs_to_launch) + "\n")
                runfile.write("process=$! \n")
                runfile.write(
                    "# Comment the following line if you don't want esm_runscripts to restart:\n"
                )
                runfile.write(
                    "cd " + config["general"]["experiment_scripts_dir"] + "\n"
                )
                runfile.write(observe_call + "\n")
                runfile.write("\n")
                runfile.write(self.append_done_statement(config, subjob) + "\n")

            runfile.write("\n")
            runfile.write("wait\n")

        config["general"]["submit_command"] = batch_system.get_submit_command(
            config, batch_or_shell, runfilename
        )

        if batch_or_shell == "shell":
            runfilestats = os.stat(runfilename)
            os.chmod(runfilename, runfilestats.st_mode | stat.S_IEXEC)

        if config["general"]["verbose"]:
            print("\n", 40 * "+ ")
            print("Contents of ", runfilename, ":")
            with open(runfilename, "r") as fin:
                print(fin.read())
            if os.path.isfile(self.bs.filename):
                print("\n", 40 * "+ ")
                print("Contents of ", self.bs.filename, ":")
                with open(self.bs.filename, "r") as fin:
                    print(fin.read())

        return config

    @staticmethod
    def write_env(config, environment=[], runfilename=""):
        # Ensures that the config is not modified in this method
        local_config = copy.deepcopy(config)
        # Allows to run it from a recipe (i.e. before a preprocessing job)
        if len(environment) == 0 or len(runfilename) == 0:
            runfilename = batch_system.get_run_filename(local_config, "prepcompute")
            environment = batch_system.get_environment(
                local_config, "prepcompute_general"
            )

        folder = local_config["general"]["thisrun_scripts_dir"]
        this_batch_system = local_config["computer"]
        runfilename_short = runfilename.split("/")[-1]
        envfilename = folder + "/env.sh"

        with open(envfilename, "w") as envfile:
            if "sh_interpreter" in this_batch_system:
                envfile.write("#!" + this_batch_system["sh_interpreter"] + "\n")
            envfile.write(f"# ENVIRONMENT used in {runfilename_short}\n")
            envfile.write("# Use this file to source the environment in your\n")
            envfile.write("# preprocessing or postprocessing scripts\n\n")
            for line in environment:
                envfile.write(line + "\n")

        return config

    @staticmethod
    def find_openmp(config):
        """
        Defines the ``heterogeneous_parallelization`` variable based on the
        ``omp_num_threads`` found in the model's sections. If any
        ``omp_num_threads`` > 1, then ``heterogeneous_parallelization`` becomes
        ``True``. Otherwise, is set to ``False``. The user has no control on setting
        this variable, as the user's choice is overridden here. This is because the
        functionality triggered by ``heterogeneous_parallelization`` is entirely
        dependent on the values of ``omp_num_threads``, so for the user, it doesn't
        make sense to define two variables for the same thing. One could think then
        that there is not need for such a variable, however, there are instances
        in which the yaml files need to know whether the simulation is heterogeneously
        parallelize (i.e. in the machine files to define some environment variables
        under a ``choose_computer.heterogeneous_parallelization``), so this is a way
        of not having to check every ``omp_num_threads`` in the yamls to verify such
        condition.

        Parameters
        ----------
        config : dict
            Dictionary containing the information about the experiment.

        Returns
        -------
        config : dict
            Dictionary containing the information about the experiment.
        """
        if config["computer"].get("heterogeneous_parallelization", False):
            user_error(
                "heterogeneous_parallelization variable",
                (
                    "Since version 6.0, ``heterogeneous_parallelization`` variable "
                    "defined by the user is ignored, and instead its value is "
                    "set to true if any ``omp_num_threads`` in the model's sections "
                    "is larger than 1. To get rid of this error, remove "
                    "``heterogeneous_parallelization`` from your yaml files. "
                    "``heterogeneous_parallelization`` can still be used from a "
                    "``choose_`` block to decice the case."
                ),
            )
        # Set ``heterogeneous_parallelization`` false, overriding whatever the user
        # has defined for this variable to be
        config["computer"]["heterogeneous_parallelization"] = False
        # Loop through the components to find ``omp_num_threads``
        for model in config:
            # Set omp_num_threads 1 if it is not defined in the model
            omp_num_threads = str(config[model].get("omp_num_threads", 1))
            # Resolve the variable if necessary
            if "${" in omp_num_threads:
                # Resolve necessary chooses and the ``omp_num_threads`` variable
                config2 = copy.deepcopy(config)
                config2 = prepare.resolve_some_choose_blocks(config2)
                omp_num_threads = find_variable(
                    [model, "omp_num_threads"],
                    config2[model]["omp_num_threads"],
                    config2,
                    [],
                    True,
                )
            # Set ``heterogeneous_parallelization`` true if needed
            if int(omp_num_threads) > 1 and model in config["general"].get(
                "valid_model_names", []
            ):
                config["computer"][
                    "heterogeneous_parallelization"
                ] = True
                if (
                    not config[model].get("nproc", False)
                    and not config[model].get("nproca", False)
                    and not config[model].get("nprocar", False)
                ):
                    config[model]["nproc"] = 1
        return config

    def het_par_launcher_lines(self, config, cluster):
        """
        Loops through the components to generate job launcher flags and execution
        commands, to be appended in substitution to the ``@components@`` tag, in
        ``computer.execution_command``, that would later be used in the writing of the
        ``.run`` file, in ``batch_system.py``.

        Parameters
        ----------
        config : dict
            Configuration dictionary containing information about the experiment and
            experiment directory.
        cluster : str
            Type of job cluster.
        """
        component_lines = []
        # Read in the separator to be used in between component calls in the job
        # launcher
        sep = config["computer"].get("launcher_comp_sep", "\\\n    ") + " "
        # Loop through the components
        for model in config["general"]["valid_model_names"]:
            command = None
            # Read in execution command
            if "execution_command" in config[model]:
                command = config[model]["execution_command"]
            elif "executable" in config[model]:
                command = config[model]["executable"]
            # Prepare the MPMD commands

# kh 24.06.22 observed behavior was:
#model:  oasis3mct
#command:  None

#model:  recom
#command:  None

#model:  jsbach
#command:  None

#model:  hdmodel
#command:  NONE

# kh 24.06.22 workaround: filter hdmodel
            if command and (command != "NONE"):
                launcher = config["computer"].get("launcher")
                launcher_flags = self.calc_launcher_flags(config, model, cluster)
                component_lines.append(f"{launcher_flags} ./{command} ")

        # Merge each component flags and commands into a single string
        components = sep.join(component_lines)
        # Substitute the ``@components@`` tag for the final launch command
        config["computer"]["execution_command"] = config["computer"][
            "execution_command"
        ].replace("@components@", components).replace("@jobtype@", cluster)

    @staticmethod
    def calc_launcher_flags(config, model, cluster):
        """
        Calculates the launcher flags for the job luncher based on the ``nproc`` of the
        different components, with the possibility of using heterogeneous
        parallelization (using both MPI and OpenMP).

        Launcher flags will vary from job launcher to job launcher, and therefore, a
        variable ``launcher_flags`` needs to be defined for each different launcher in
        the ``pbs.yaml`` configuration file. This variable should contain the flags
        themselves followed by their respective values to be calculated in this method
        and substituted in their respective tags of the ``launcher_flags`` variable.
        This method supports the following variables:

            - ``@nproc@``: total number of MPI tasks for the component taken from
              ``<component>.nproc``.
            - ``@nproc_per_node@``: number of MPI tasks per node calculated from
              the number of number of CPUs per MPI task (``<component>.cpus_per_proc``),
              and the number of cores per node in the machine
              (``computer.cores_per_node``).
            - ``@cpus_per_proc@``: number of CPUs per MPI task to be used
              (``<component>.cpus_per_proc``). If not defined, it takes the value of
              ``<component>.omp_num_threads``. This variables are not necessarily the
              same, to allow the use of  more cores per MPI task than OpenMP threads,
              if necessary.
            - ``@omp_num_threads@``: number of OpenMP threads per MPI task
              (``<component>.omp_num_threads``). Its default value is ``1``, if not
              defined. The number of OpenMP threads per task cannot be larger than the
              number of CPUs per task.

        Parameters
        ----------
        config : dict
            Configuration dictionary containing information about the experiment and
            experiment directory.
        model : str
            Component for which the flags are to be calculated.
        cluster : str
            Type of job cluster.

        Returns
        -------
        launcher_flags : str
            Launcher flags string with the calculated numbers for the ``model`` already
            substituted in the tags.
        """
        launcher = config["computer"]["launcher"]
        launcher_flags = config["computer"]["launcher_flags_per_component"]
        # Cores per node
        # cores_per_node = config["computer"]["cores_per_node"]
        if cluster == "compute":
            cores_per_node = config["computer"]["partitions"]["compute"][
                "cores_per_node"
            ]
        else:
            cores_per_node = config["computer"]["partitions"]["pp"][
                "cores_per_node"
            ]
        # Get the OMP number of threads
        omp_num_threads = config[model].get("omp_num_threads", 1)

        if "nproc" in config[model]:
            # aprun flags commented following the conventions in p. 14 of the ALEPH ppt
            # manual day_1.session_2.advanced_use_of_aprun.ppt

            # Total number of PEs (MPI-ranks) (e.g. aprun -n)
            nproc = config[model]["nproc"]
            # CPUs per MPI-rank (e.g. aprun -d)
            cpus_per_proc = config[model].get("cpus_per_proc", omp_num_threads)
            # Check for CPUs and OpenMP threads
            if omp_num_threads > cpus_per_proc:
                esm_parser.user_error(
                    "OpenMP configuration",
                    (
                        "The number of OpenMP threads cannot be larger than the number"
                        + "of CPUs per MPI task requested. Your values:\n"
                        + f"    {model}.omp_num_threads: {omp_num_threads}\n"
                        + f"    {model}.cpus_per_proc: {cpus_per_proc}\n"
                    ),
                )
        elif "nproca" in config[model] and "nprocb" in config[model]:
            # ``nproca``/``nprocb`` not compatible with ``omp_num_threads``
            if omp_num_threads > 1:
                esm_parser.user_note(
                    "nproc",
                    "``nproca``/``nprocb`` not compatible with ``omp_num_threads``",
                )
            nproc = config[model]["nproca"] * config[model]["nprocb"]
            cpus_per_proc = 1
            omp_num_threads = 1
        else:

# kh 22.06.22 defensive (user_error/user_note could also be added here)
            nproc = 0
            cpus_per_proc = 0
#           omp_num_threads = 0

        # Number of nodes needed
        nodes = int(nproc * cpus_per_proc / cores_per_node) + (
            (nproc * cpus_per_proc) % cores_per_node > 0
        )
        config[model]["nodes"] = nodes

        # PEs (MPI-ranks) per compute node (e.g. aprun -N)
        nproc_per_node = int(nproc / nodes)

        # Replace tags in the laucher flags
        replacement_tags = [
            ("@nnodes@", nodes),
            ("@nproc@", nproc),
            ("@nproc_per_node@", nproc_per_node),
            ("@cpus_per_proc@", cpus_per_proc),
            ("@omp_num_threads@", omp_num_threads),
        ]
        # Replace all tags
        for (tag, repl) in replacement_tags:
            launcher_flags = launcher_flags.replace(tag, str(repl))
        # Substitute @MODEL@ with the model name
        launcher_flags = launcher_flags.replace("@MODEL@", model.upper())

        return launcher_flags



def submits_another_job(config, cluster):
    clusterconf = config["general"]["workflow"]["subjob_clusters"][cluster]
    if clusterconf.get("next_submit", []) == []:
        return False
    return True



from . import database
from datetime import datetime
import sqlalchemy


def database_entry(config):
    if config.get("general", {}).get("use_database", True):
        if config["general"]["check"]:
            database_entry_check(config)
        elif config["general"]["jobtype"] == "compute":
            database_entry_start(config)
    return config


def database_basic_entry(config):
    thisrun = (
        database.session.query(database.experiment)
        .filter_by(expid=config["general"]["expid"])
        .filter_by(run_timestamp=config["general"]["run_datestamp"])
        .all()
    )

    if thisrun == []:
        thisrun = database.experiment(
            expid=config["general"]["expid"],
            setup_name=config["general"]["setup_name"],
            run_timestamp=config["general"]["run_datestamp"],
            timestamp=datetime.now(),
            outcome="None",
            exp_folder=f"{config['general']['base_dir']}/{config['general']['expid']}/",
        )
        database.session.add(thisrun)
    else:
        thisrun = thisrun[-1]
        thisrun.timestamp = datetime.now()
    return thisrun


def try_to_commit():
    try:
        database.session.commit()
    except sqlalchemy.exc.OperationalError as e:
        print("Sorry, there was some SQL Error!")
        print(e)


def database_entry_check(config):
    thisrun = database_basic_entry(config)
    thisrun.outcome = "check"
    try_to_commit()


def database_entry_start(config):
    thisrun = database_basic_entry(config)
    thisrun.outcome = "started"
    try_to_commit()


def database_entry_success(config):
    thisrun = database_basic_entry(config)
    thisrun.outcome = "success"
    try_to_commit()


def database_entry_crashed(config):
    thisrun = database_basic_entry(config)
    thisrun.outcome = "crashed"
    try_to_commit()



import os
import time
import subprocess
import copy

import f90nml
import yaml
import stat

import esm_calendar
import esm_parser
import esm_runscripts

from .batch_system import batch_system
from .filelists import copy_files, log_used_files
from .helpers import evaluate 
from .namelists import Namelist

#####################################################################
#                                   compute jobs                    #
#####################################################################


def run_job(config):
    config["general"]["relevant_filetypes"] = [
        "bin",
        "config",
        "forcing",
        "input",
        "restart_in",
    ]
    config = evaluate(config, "prepcompute", "prepcompute_recipe")
    return config


def compile_model(config):
    """Compiles the desired model before the run starts"""
    model = config["general"]["setup_name"]
    version = config["general"].get("version") or config[model].get("version")
    if not version:
        return config
    if config.get("general", {}).get("run_number") == 1:
        print("First year, checking if we need to compile...")
        if not config.get("general", {}).get("use_compiled_model", True):
            print(f"Huhu --> compiling {model}-{version}")
            subprocess.run(
                f"esm_master install-{model}-{version}",
                shell=True,
                cwd=config["general"]["experiment_src_dir"],
            )
            config["general"]["model_dir"] = (
                config["general"]["experiment_src_dir"] + f"/{model}-{version}"
            )
    return config


def all_files_to_copy_append(
    config, model, filetype, categ, file_source, file_interm, file_target
):
    if file_source:
        if not filetype + "_sources" in config[model]:
            config[model][filetype + "_sources"] = {}
        config[model][filetype + "_sources"][categ] = file_source
    if file_interm:
        if not filetype + "_intermediate" in config[model]:
            config[model][filetype + "_intermediate"] = {}
        config[model][filetype + "_intermediate"][categ] = file_interm
    if file_target:
        if (
            filetype in config["general"]["in_filetypes"]
            and filetype + "_in_work" in config[model]
        ):
            config[model][filetype + "_in_work"][categ] = file_target
        else:
            # print (filetype)
            # print (file_target)
            # print (categ)
            # print (config["general"]["out_filetypes"])
            if not filetype + "_targets" in config[model]:
                config[model][filetype + "_targets"] = {}
            config[model][filetype + "_targets"][categ] = file_target

    if filetype + "_files" in config[model]:
        config[model][filetype + "_files"][categ] = categ

    return config


def prepare_coupler_files(config):
    if config["general"]["standalone"] is False:
        coupler_filename = config["general"]["coupler"].prepare(
            config, config["general"]["coupler_config_dir"]
        )
        coupler_name = config["general"]["coupler"].name
        if coupler_name == "yac":
            couplingfile = "coupling.xml"
        else:
            couplingfile = "namcouple"

        all_files_to_copy_append(
            config,
            coupler_name,
            "config",
            couplingfile,
            config["general"]["coupler_config_dir"] + "/" + coupler_filename,
            None,
            None,
        )
    return config


def create_empty_folders(config):
    for model in list(config):
        if "create_folders" in config[model]:
            folders = config[model]["create_folders"]
            if not type(folders) == list:
                folders = [folders]
            for folder in folders:
                if not os.path.isdir(folder):
                    os.mkdir(folder)
    return config


def create_new_files(config):
    for model in list(config):
        for filetype in config["general"]["all_filetypes"]:
            if "create_" + filetype in config[model]:
                filenames = config[model]["create_" + filetype].keys()

                for filename in filenames:

                    full_filename = (
                        config[model]["thisrun_" + filetype + "_dir"] + "/" + filename
                    )
                    if not os.path.isdir(os.path.dirname(full_filename)):
                        os.mkdir(os.path.dirname(full_filename))
                    with open(
                        full_filename,
                        "w",
                    ) as createfile:
                        actionlist = config[model]["create_" + filetype][filename]
                        for action in actionlist:
                            if "<--append--" in action:
                                appendtext = action.replace("<--append--", "")
                                createfile.write(appendtext.strip() + "\n")
                    # make executable, just in case
                    filestats = os.stat(full_filename)
                    os.chmod(full_filename, filestats.st_mode | stat.S_IEXEC)

                    all_files_to_copy_append(
                        config,
                        model,
                        filetype,
                        filename,
                        config[model]["thisrun_" + filetype + "_dir"] + "/" + filename,
                        filename,
                        filename,
                    )
    return config


def modify_files(config):
    # for model in config:
    #     for filetype in config["general"]["all_model_filetypes"]:
    #         if filetype == "restart":
    #             nothing = "nothing"
    return config


def modify_namelists(config):
    # Load and modify namelists:

    if config["general"]["verbose"]:
        print("\n" "- Setting up namelists for this run...")
        for index, model in enumerate(config["general"]["valid_model_names"]):
            print(f'{index+1}) {config[model]["model"]}')
        print()

    for model in config["general"]["valid_model_names"]:
        config[model] = Namelist.nmls_load(config[model])
        config[model] = Namelist.nmls_remove(config[model])
        if model == "echam":
            config = Namelist.apply_echam_disturbance(config)
            config = Namelist.echam_transient_forcing(config)
        if model == "fesom":
            config = Namelist.apply_iceberg_calving(config)
        config[model] = Namelist.nmls_modify(config[model])
        config[model] = Namelist.nmls_finalize(
            config[model], config["general"]["verbose"]
        )

    if config["general"]["verbose"]:
        print("::: end of namelist section\n")
    return config


def copy_files_to_thisrun(config):
    if config["general"]["verbose"]:
        print("PREPARING EXPERIMENT")
        # Copy files:
        print("\n" "- File lists populated, proceeding with copy...")
        print("- Note that you can see your file lists in the config folder")
        print("- You will be informed about missing files")

    counter = 0
    count_max = 90
    if (
        config["general"].get("iterative_coupling", False)
        and config["general"]["chunk_number"] > 1
    ):
        if "files_to_wait_for" in config["general"]:
            for file_base in config['general'].get('files_to_wait_for'):
                file = os.path.join(config['general']['experiment_couple_dir'], file_base)
                while counter < count_max:
                    counter = counter + 1
                    if os.path.isfile(file):
                        print("File found: ", file)
                        break
                    else:
                        print("Waiting for file: ", file)
                        print("Sleep for 10 seconds...")
                        time.sleep(10)

    # MA: TODO: this should go somewhere else, maybe on its on module and then inserted on a recipe
    if "fesom" in config["general"]["valid_model_names"]:
        if config["fesom"].get("use_icebergs", False) and config["fesom"].get("use_icesheet_coupling", False):
            if config["general"].get("run_number", 0) == 1:
                if not os.path.isfile(
                    config["general"]["experiment_couple_dir"] + "/num_non_melted_icb_file"
                ):
                    with open(config["general"]["experiment_couple_dir"] + "/num_non_melted_icb_file", "w") as f:
                        f.write("0")
            else:
                num_lines = sum(1 for line in open(os.path.join(config["fesom"]["experiment_restart_in_dir"], "iceberg.restart.ISM")))
                with open(config["general"]["experiment_couple_dir"] + "/num_non_melted_icb_file", "w") as f:
                    f.write(str(num_lines))

    log_used_files(config)

    config = copy_files(
        config, config["general"]["in_filetypes"], source="init", target="thisrun"
    )
    return config


def copy_files_to_work(config):
    if config["general"]["verbose"]:
        print("PREPARING WORK FOLDER")
    config = copy_files(
        config, config["general"]["in_filetypes"], source="thisrun", target="work"
    )
    return config


def _write_finalized_config(config):
    """Writes <expid>_finished_config.yaml file
    Parameters
    ----------
    config : esm-tools config object
    """
    # first define the representers for the non-built-in types, as recommended
    # here: https://pyyaml.org/wiki/PyYAMLDocumentation
    def date_representer(dumper, date):
        return dumper.represent_str(f"{date.output()}")

    def calendar_representer(dumper, calendar):
        # Calendar has a __str__ method
        return dumper.represent_str(str(calendar))

    def batch_system_representer(dumper, batch_system):
        return dumper.represent_str(f"{batch_system.name}")

    def coupler_representer(dumper, coupler):
        # prevent dumping of whole namcouple
        return dumper.represent_str(f"{coupler.name}")

    def oasis_representer(dumper, oasis):
        return dumper.represent_str(f"{oasis.name}")

    def namelist_representer(dumper, f90nml):
        return dumper.represent_str(f"f90nml.name")

    # dumper object for the ESM-Tools configuration
    class EsmConfigDumper(yaml.dumper.Dumper):
        pass

    # pyyaml does not support tuple and prints !!python/tuple
    EsmConfigDumper.add_representer(
        tuple, yaml.representer.SafeRepresenter.represent_list
    )

    # Determine how non-built-in types will be printed be the YAML dumper
    EsmConfigDumper.add_representer(esm_calendar.Date, date_representer)

    EsmConfigDumper.add_representer(
        esm_calendar.esm_calendar.Calendar, calendar_representer
    )
    # yaml.representer.SafeRepresenter.represent_str)

    EsmConfigDumper.add_representer(
        esm_parser.esm_parser.ConfigSetup,
        yaml.representer.SafeRepresenter.represent_dict,
    )

    EsmConfigDumper.add_representer(batch_system, batch_system_representer)

    # format for the other ESM data structures
    EsmConfigDumper.add_representer(
        esm_runscripts.coupler.coupler_class, coupler_representer
    )

    EsmConfigDumper.add_representer(
        f90nml.namelist.Namelist, namelist_representer
    )

    if "oasis3mct" in config:
        EsmConfigDumper.add_representer(esm_runscripts.oasis.oasis, oasis_representer)

    thisrun_config_dir = config["general"]["thisrun_config_dir"]
    expid = config["general"]["expid"]
    it_coupled_model_name = config["general"]["iterative_coupled_model"]
    config_file_path = (
        f"{thisrun_config_dir}/"
        f"{expid}_{it_coupled_model_name}finished_config.yaml"
    )
    with open(config_file_path, "w") as config_file:
        # Avoid saving ``prev_run`` information in the config file
        config_final = copy.deepcopy(config)  # PrevRunInfo
        del config_final["prev_run"]  # PrevRunInfo

        out = yaml.dump(
            config_final, Dumper=EsmConfigDumper, width=10000, indent=4
        )  # PrevRunInfo
        config_file.write(out)
    return config


def _show_simulation_info(config):
    print()
    print(80 * "=")
    print("STARTING SIMULATION JOB!")
    print(f"Experiment ID = {config['general']['expid']}")
    print(f"Setup = {config['general']['setup_name']}")
    if "coupled_setup" in config["general"]:
        print("This setup consists of:")
        for model in config["general"]["valid_model_names"]:
            print(f"- {model}")
    print("Experiment is installed in:")
    print(f"       {config['general']['base_dir']}/{config['general']['expid']}")
    print(80 * "=")
    print()
    return config



"""
=================================
ESM Tools (Package Documentation)
=================================

This package contains almost no code, but instead is where most of the YAML
configurations files are distributed. Default namelists are included as well as
example runscripts. This section **only** documents the code contained in the
module, please refer to the handbook for user documentation as well as API
documentation for the various sub-modules of the project.

Accessing Configuration
-----------------------

To access a particular configuration, you can use::

    >>> from esm_tools import read_config_file
    >>> ollie_config = read_config_file("machines/ollie")

Important note here is that the configuration file **has not yet been parsed**,
so it's just the dictionary representation of the YAML.
"""

__author__ = """Dirk Barbi, Paul Gierz"""
__email__ = "dirk.barbi@awi.de"
__version__ = "6.17.5"

import functools
import inspect
import operator
import os
import pathlib
import shutil
import site
import sys

import pkg_resources
import yaml
from loguru import logger

# Setup Loguru for the following cases:
# A) If user sets
if os.environ.get("DEBUG_ESM_TOOLS"):
    logger.add(sys.stderr, filter={"": "WARNING", "esm_tools": "DEBUG"})
elif os.environ.get("CI"):
    logger.remove()
    logger.add(sys.stderr, filter={"": "WARNING"})
    logger.add(
        "out.log",
        filter={"": "WARNING", "esm_tools": "DEBUG"},
        backtrace=True,
        diagnose=True,
    )
elif os.environ.get("CI_VERBOSE"):
    logger.remove()
    logger.add(sys.stderr, filter={"": "WARNING", "esm_tools": "DEBUG"})
    logger.add(
        "out.log",
        filter={"": "WARNING", "esm_tools": "DEBUG"},
        backtrace=True,
        diagnose=True,
    )
else:
    logger.remove()


def caller_wrapper(func):
    @functools.wraps(func)
    def wrapped_func(*args, **kwargs):
        logger.debug(f"{inspect.stack()[1].function} --> {func.__name__}")
        return func(*args, **kwargs)

    return wrapped_func


def _transform(nested_list):
    """Transform irregular 2D list into a regular one."""
    regular_list = []
    for ele in nested_list:
        if type(ele) is list:
            regular_list.append(ele)
        else:
            regular_list.append([ele])
    return regular_list


@caller_wrapper
def _get_real_dir_from_pth_file(subfolder):
    logger.debug(f"Trying to resolve: {subfolder}")
    if subfolder.startswith("/"):
        logger.warning("Subfolder is strange!")
        logger.warning(subfolder)
    site_packages_dirs = functools.reduce(
        operator.iconcat,
        _transform([site.getusersitepackages(), site.getsitepackages()]),
        [],
    )
    logger.debug(site_packages_dirs)
    for site_package_dir in site_packages_dirs:
        logger.debug(f"Working on {site_package_dir}")
        # Read the pth file:
        if pathlib.Path(f"{site_package_dir}/esm-tools.egg-link").exists():
            with open(f"{site_package_dir}/esm-tools.egg-link", "r") as f:
                paths = [p.strip() for p in f.readlines()]
            # NOTE(PG): a pathlib.Path has a method resolve, which removes
            # things like "foo/baz/../bar" in the path to "foo/bar"
            logger.debug(f"paths={paths}")
            logger.debug(f"subfolder={subfolder}")
            actual_package_data_dir = pathlib.Path(
                f"{paths[0]}/{paths[1]}/{subfolder}/"
            )
            logger.debug("Before resolve:")
            logger.debug(actual_package_data_dir)
            logger.debug("After resolve:")
            actual_package_data_dir = actual_package_data_dir.resolve()
            logger.debug(actual_package_data_dir)
            try:
                assert actual_package_data_dir.exists()
            # NOTE(PG): there is probably a better way of doing that than with assert.
            except AssertionError as e:
                logger.debug(
                    f"Assumed path {actual_package_data_dir} did not exist! We tried:"
                )
                logger.debug(f"paths[0]={paths[0]}")
                logger.debug(f"paths[1]={paths[1]}")
                if paths[1].startswith("/"):
                    logger.debug(
                        f"{paths[1]} starts with a slash, assuming absolute path!"
                    )
                    actual_package_data_dir = pathlib.Path(
                        f"{paths[1]}/{subfolder}/"
                    ).resolve()
                    try:
                        assert actual_package_data_dir.exists()
                    except AssertionError as e:
                        logger.error("Could not determine path!")
                        break  # Break out of the for loop
            logger.debug(f"actual_package_data_dir={actual_package_data_dir}")
            return actual_package_data_dir
    raise FileNotFoundError(
        f"Could not determine where {subfolder}'s path is inside the esm-tools installation! These were searched for info: {site_packages_dirs}"
    )


def _get_namelist_filepath_standard_install(namelist):
    return pkg_resources.resource_filename("esm_tools.namelists", namelist)


def _get_namelist_filepath_editable_install(namelist):
    return f"{_get_real_dir_from_pth_file('namelists')}/{namelist}"


def _get_config_filepath_standard_install(config):
    return pkg_resources.resource_filename("esm_tools.configs", config)


def _get_config_filepath_editable_install(config):
    return f"{_get_real_dir_from_pth_file('configs')}/{config}"


def _get_runscript_filepath_standard_install(runscript):
    return pkg_resources.resource_filename("esm_tools.runscripts", runscript)


def _get_runscript_filepath_editable_install(runscript):
    return f"{_get_real_dir_from_pth_file('runscripts')}/{runscript}"


def get_config_as_str(config):
    return pkg_resources.resource_string("esm_tools.configs", config)


def _list_config_dir_standard_install(dir_path):
    return pkg_resources.resource_listdir("esm_tools.configs", dir_path)


def _list_config_dir_editable_install(dir_path):
    return os.listdir(f"{_get_real_dir_from_pth_file('configs')}/{dir_path}")


# For more information on how this works, see here:
# https://stackoverflow.com/questions/62550952/including-package-data-python-from-top-level-when-package-is-in-subdirectory/62552188#62552188
def _read_config_standard_install(config):
    """
    Reads a config file for the case that you have done a standard pip install.

    Parameters
    ----------
    config : str
        The configuration to read, e.g. "machines/ollie.yaml"

    Returns
    -------
    dict :
        A dictionary representation of the configuration
    """
    if not config.endswith(".yaml"):
        config += ".yaml"
    configstr = pkg_resources.resource_string("esm_tools.configs", config)
    # configstr = pkg_resources.resource_string("configs", config)
    configdict = yaml.load(configstr, Loader=yaml.FullLoader)
    return configdict


def _read_namelist_standard_install(nml):
    """
    Reads a namelist file for the case that you have done a standard pip install.

    Parameters
    ----------
    nml : str
        The nml to read, e.g. "echam/6.3.04p2/PALEO/namelist.echam"

    Returns
    -------
    str :
        A string representation of the namelist file. This should later be
        passed to the f90nml package.
    """
    return pkg_resources.resource_string("esm_tools.namelists", nml)


def _read_config_editable_install(config):
    """
    Reads a config file for the case you have done an editable/develop install
    (e.g. pip -e).

    Parameters
    ----------
    dict :
        A dictionary representation of the configuration
    """
    if not config.endswith(".yaml"):
        config += ".yaml"
    # Note the only difference here is apparently swapping out
    # esm_tools.configs for just configs. Not sure how that works, but it seems
    # to be fine...
    with open(_get_config_filepath_editable_install(config), "r") as cfg:
        configstr = cfg.read()
    configdict = yaml.load(configstr, Loader=yaml.FullLoader)
    return configdict


def _copy_config_folder_standard_install(dest_path):
    src_path = pkg_resources.resource_filename("esm_tools.configs", ".")
    return shutil.copytree(src_path, dest_path)


def _copy_config_folder_editable_install(dest_path):
    src_path = _get_config_filepath_editable_install("")
    return shutil.copytree(src_path, dest_path)


def _copy_namelist_folder_standard_install(dest_path):
    src_path = pkg_resources.resource_filename("esm_tools.namelists", ".")
    return shutil.copytree(src_path, dest_path)


def _copy_namelist_folder_editable_install(dest_path):
    src_path = _get_namelist_filepath_editable_install("")
    return shutil.copytree(src_path, dest_path)


def _copy_runscript_folder_standard_install(dest_path):
    src_path = pkg_resources.resource_filename("esm_tools.runscripts", ".")
    return shutil.copytree(src_path, dest_path)


def _copy_runscript_folder_editable_install(dest_path):
    src_path = _get_runscript_filepath_editable_install("")
    return shutil.copytree(src_path, dest_path)


def _read_namelist_editable_install(nml):
    """
    Reads a namelist file for the case that you have done an editable/develop install.

    Parameters
    ----------
    nml : str
        The nml to read, e.g. "echam/6.3.04p2/PALEO/namelist.echam"

    Returns
    -------
    str :
        A string representation of the namelist file. This should later be
        passed to the f90nml package.
    """
    with open(_get_namelist_filepath_editable_install("namelists", nml), "r") as nml:
        return nml.read()


# PG: Blatant theft:
# https://stackoverflow.com/questions/42582801/check-whether-a-python-package-has-been-installed-in-editable-egg-link-mode
def _dist_is_editable(dist):
    """Is distribution an editable install?"""
    for path_item in sys.path:
        egg_link = os.path.join(path_item, dist.replace("_", "-") + ".egg-link")
        if os.path.isfile(egg_link):
            return True
    return False


EDITABLE_INSTALL = _dist_is_editable("esm_tools")
"""
bool : Shows if the installation is installed in editable mode or not.
"""


# Create a function that "does the right thing", depending on how this module
# is installed
@caller_wrapper
def read_config_file(config):
    """
    Reads a configuration file, which should be seperated by "/". For example,
    "machines/ollie" will retrieve the (unparsed) configuration of the Ollie
    supercomputer.

    Parameters
    ----------
    config : str
        Configuration to get, e.g. machines/ollie.yaml, or echam/echam. You may
        omit the ".yaml" ending if you want, it will be appended automatically
        if not already there.

    Returns
    -------
    dict :
        A dictionary representation of the config.
    """
    if EDITABLE_INSTALL:
        return _read_config_editable_install(config)
    return _read_config_standard_install(config)


@caller_wrapper
def list_config_dir(dirpath):
    if EDITABLE_INSTALL:
        return _list_config_dir_editable_install(dirpath)
    return _list_config_dir_standard_install(dirpath)


@caller_wrapper
def copy_config_folder(dest_path):
    if EDITABLE_INSTALL:
        return _copy_config_folder_editable_install(dest_path)
    return _copy_config_folder_standard_install(dest_path)


@caller_wrapper
def copy_namelist_folder(dest_path):
    if EDITABLE_INSTALL:
        return _copy_namelist_folder_editable_install(dest_path)
    return _copy_namelist_folder_standard_install(dest_path)


@caller_wrapper
def copy_runscript_folder(dest_path):
    if EDITABLE_INSTALL:
        return _copy_runscript_folder_editable_install(dest_path)
    return _copy_runscript_folder_standard_install(dest_path)


@caller_wrapper
def get_namelist_filepath(namelist=""):
    if EDITABLE_INSTALL:
        return _get_namelist_filepath_editable_install(namelist)
    return _get_namelist_filepath_standard_install(namelist)


@caller_wrapper
def get_config_filepath(config=""):
    if EDITABLE_INSTALL:
        cpath = _get_config_filepath_editable_install(config)
    else:
        cpath = _get_config_filepath_standard_install(config)
    return cpath


@caller_wrapper
def get_runscript_filepath(runscript=""):
    if EDITABLE_INSTALL:
        return _get_runscript_filepath_editable_install(runscript)
    return _get_runscript_filepath_standard_install(runscript)


@caller_wrapper
def read_namelist_file(nml):
    """Reads a namelist file from a path, seperated by "/". Similar to ``read_config_file``

    Parameters
    ----------
    nml : str
        The namelist to load

    Returns
    -------
    str :
        A string of the namelist file
    """
    if EDITABLE_INSTALL:
        return _read_namelist_editable_install(nml)
    return _read_namelist_standard_install(nml)



from sqlalchemy import create_engine, Column, Integer, String, Sequence, DateTime
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

import os

main_database_file = os.path.expanduser("~") + "/.esm_tools/esmtools.db"
if not os.path.isdir(os.path.expanduser("~") + "/.esm_tools"):
    os.mkdir(os.path.expanduser("~") + "/.esm_tools")

engine = create_engine("sqlite:///" + main_database_file, echo=False)
base = declarative_base()


class database_location(base):
    __tablename__ = "locations"

    id = Column(Integer, Sequence("user_id_seq"), primary_key=True)
    table_name = Column(String)
    location = Column(String)
    class_in = Column(String)

    @staticmethod
    def topline():
        print()
        print("Databases:")
        print(
            "{0: >4}".format("ID")
            + "   "
            + "{0: >17}".format("table_name")
            + "   "
            + "{0: >60}".format("location")
            + "   "
            + "{0: >17}".format("class_in")
        )

    def __repr__(self):
        return "%4s   %17s   %85s   %17s" % (
            str(self.id),
            str(self.table_name),
            str(self.location),
            str(self.class_in),
        )


base.metadata.create_all(engine)
connection = engine.connect()
Session = sessionmaker(bind=engine)
Session.configure(bind=engine)
session = Session()

# close database connection
session.close()
connection.close()
engine.dispose()


def register(table_name, given_location, class_in):
    loc_id = 0
    action = "new"
    if os.path.isfile(main_database_file):
        all_locations = session.query(database_location).all()

        for location in all_locations:
            if location.table_name == table_name:
                if location.location == given_location:
                    action = "skip"
                    break
                else:
                    loc_id = location.id
                    action = "update"
                    break

    if action == "new":
        new_loc = database_location(
            table_name=table_name, location=given_location, class_in=class_in
        )
        session.add(new_loc)
        session.commit()
    elif action == "update":
        loc = session.query(database_location).filter_by(id=loc_id).first()
        loc.location = given_location
        session.commit()



import os, sys, _io
from . import helpers


def initialize_logfiles(config, org_jobtype):
    global logfile_handle
    global logfile_run_number

    logfile_run_number = str(config["general"]["run_number"])

    log_stuff = False
    if os.path.isdir(os.path.dirname(config["general"]["experiment_log_file"])):
        if not org_jobtype == "inspect":
            log_stuff = True

    config = set_logfile_name(config, "")

    if log_stuff:

        helpers.write_to_log(
            config,
            [
                org_jobtype,
                logfile_run_number,
                str(config["general"]["current_date"]),
                str(config["general"]["jobid"]),
                "- start",
            ],
        )
        logfile = RuntimeLogger(
            config["general"]["logfile_path"],
            "w",
            buffering=1,
        )
    else:
        logfile = sys.stdout

    logfile_handle = logfile
    return config


def finalize_logfiles(config, org_jobtype):

    if os.path.isdir(os.path.dirname(config["general"]["experiment_log_file"])):
        log_stuff = True

    if log_stuff:
        helpers.write_to_log(
            config,
            [
                org_jobtype,
                logfile_run_number,
                str(config["general"]["current_date"]),
                str(config["general"]["jobid"]),
                "- done",
            ],
        )

    if not logfile_handle == sys.stdout:
        logfile_handle.close()
    return config


def set_logfile_name(config, jobtype=None):

    if not jobtype:
        jobtype = config["general"]["jobtype"]

    filejobtype = jobtype
    # if "observe" in filejobtype:
    #    filejobtype = filejobtype.replace("observe_", "")

    # if "newrun" in filejobtype:
    #    filejobtype = config["general"]["workflow"]["subjob_clusters"][jobtype].get("next_submit")[0]

    # if filejobtype == "prepcompute":
    #    filejobtype = "compute"

    # if "_" + config["general"]["setup_name"] in filejobtype:
    #    filejobtype = filejobtype.replace("_" + config["general"]["setup_name"], "")

    # called_from = config["general"]["workflow"]["subjob_clusters"][jobtype].get("called_from", "SOMETHINgUSELESS")

    # if "_" + called_from in filejobtype:
    #    filejobtype = filejobtype.replace("_" + called_from, "")

    filename = (
        config["general"]["expid"]
        + "_"
        + config["general"]["setup_name"]
        + "_"
        + filejobtype
        + "_"
        + config["general"]["run_datestamp"]
        + ".log"
    )

    config["general"]["logfile_path"] = (
        config["general"]["experiment_log_dir"] + "/" + filename
    )

    config["general"]["logfile_path_in_run"] = (
        config["general"]["thisrun_log_dir"] + "/" + filename
    )

    # if os.path.isfile(config["general"]["logfile_path"]):
    #    if not os.path.isfile(config["general"]["logfile_path_in_run"]):
    #        os.symlink(
    #            config["general"]["logfile_path"],
    #            config["general"]["logfile_path_in_run"]
    #            )

    return config


class RuntimeLogger(_io.TextIOWrapper):
    """
    Logger object that takes care of both, writing the stdout and stderr to the
    'mini-log' file corresponding to the different elements of the workflow (observe,
    tidy, ...), and writing to the system stdout/stderr so that it is catched by SLURM
    (or I guess PBS) and put into the ``*compute*.log`` specified in the ``.run`` file.
    The ``*compute*.log`` file is then used to catch errors by ``observe``.

    .. Note: This is a fast and minimal solution for release 6.0, but all this could be done
    without needing this object, once ``loguru`` is implemented everywhere in some
    point in the future.

    .. Note: This is a subclass of ``_io.TextIOWrapper`` which is horrible! I could
    have used simply ``object`` instead, but this was the only way I (MA) found for
    ``f90nml`` package to be able to print namelists in the following lines of
    ``namelist.py``::
        def nmls_output(mconfig):
            all_nmls = {}

            for nml_name, nml_obj in mconfig.get("namelists", {}).items():
                all_nmls[nml_name] = nml_obj  # PG: or a string representation?
            for nml_name, nml in all_nmls.items():
                message = f'\nFinal Contents of {nml_name}:'
                print(message)
                print(len(message) * '-')
                nml.write(sys.stdout)  # Problematic line if the sys.stdout is not a file-like object
                print('-' * 80)
                print(f'::: end of the contents of {nml_name}\n')
            return mconfig

    Hopefully, we are not using any other methods than the ones specified here...
    """
    # FIXME(PG): This can be better.....

    def __init__(self, *args, **kwargs):
        """
        Initializes the logger object, and opens the mini-log file, to be used as a
        target for ``sys.stdout``::
            sys.stdout = LogFile(args, kwargs)

        Parameters
        ----------
        args, kwargs
            Use the same parameters as you would with the ``open`` command.
        """
        # Direct system's stderr to stdout
        sys.stderr = sys.stdout
        # Store the system's stdout as an object
        self.stdout = sys.stdout
        # Open the mini-log file
        self.file_obj = open(*args, **kwargs)

    def write(self, *args, **kwargs):
        """
        Writes messages to the mini-log file and to the system's stdout.
        """
        # Write into the mini-log file
        self.file_obj.write(*args, **kwargs)
        # Write into the system's stdout (so that it makes it to *compute*.log defined
        # in the .run file)
        self.stdout.write(args[0])

    def flush(self):
        pass

    def close(self, *args, **kwargs):
        """
        Closes the mini-log file.
        """
        # Close the mini-log file
        self.file_obj.close(*args, **kwargs)



import random
import numpy as np
import numexpr as ne
import xarray as xr
import pandas as pd
import sys
import os
from tqdm import tqdm
from datetime import datetime

RES_pism = 5
rho_ice = 910
height_icb = 250

N_min = 0
N_max = 10000

#Nlength = np.array([5000, 500, 100, 50, 20, 10, 5, 2, 1]) #icb_scaled20
Nlength = np.array([5000, 500, 100, 50, 20, 10, 1, 1, 1]) #icb_scaled21
#Nlength = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1])
#sdepth = np.array([40, 67, 133, 175, 250, 250, 250, 250, 250]) * 7/8
sdepth = np.array([height_icb]*9)
print("sdepth = ", sdepth)
slength_bins = np.array([0, 100, 200, 350, 500, 700, 900, 1200, 1600, 2200])
svol_bins = np.multiply(slength_bins, slength_bins) * height_icb
print("svol_bins = ", svol_bins)

mu_length = 250
trunc_length=0

vol_trunc=10e6 #trunc_length**3*16/21
pism_trunc_fkt=-vol_trunc/(RES_pism*RES_pism*10e6/rho_ice)

xs = []
ys = []
length = []

random.seed()

ifile = sys.argv[1]
mesh_path = sys.argv[2]
icb_path = sys.argv[3]

nod2d_file = os.path.join(mesh_path, "nod2d.out")
elem2d_file = os.path.join(mesh_path, "elem2d.out")

#read PISM file
def read_pism_file(ifile):
    fl = xr.open_dataset(ifile).squeeze()

    data = fl.tendency_of_ice_amount_due_to_discharge.where(fl.tendency_of_ice_amount_due_to_discharge < pism_trunc_fkt*N_min)
    data.to_netcdf(os.path.join(icb_path, "icb_mask.nc"))
    #data = fl.discharge_flux_cumulative.where(fl.discharge_flux_cumulative < 0, drop=True)
    data = np.array(data).reshape(1, len(data.x)*len(data.y))
    data = data[~np.isnan(data)]

    lon_bnds = fl.lon_bnds.where(fl.tendency_of_ice_amount_due_to_discharge < pism_trunc_fkt*N_min, drop=True)
    #lon_bnds = fl.lon_bnds.where(fl.discharge_flux_cumulative < 0, drop=True)
    lon_bnds = np.array(lon_bnds).reshape(len(lon_bnds.x)*len(lon_bnds.y), 4)
    lon_bnds = lon_bnds[~np.isnan(lon_bnds)]
    lon_bnds = lon_bnds.reshape(int(len(lon_bnds)/4), 4)

    lat_bnds = fl.lat_bnds.where(fl.tendency_of_ice_amount_due_to_discharge < pism_trunc_fkt*N_min, drop=True)
    #lat_bnds = fl.lat_bnds.where(fl.discharge_flux_cumulative < 0, drop=True)
    lat_bnds = np.array(lat_bnds).reshape(len(lat_bnds.x)*len(lat_bnds.y), 4)
    lat_bnds = lat_bnds[~np.isnan(lat_bnds)]
    lat_bnds = lat_bnds.reshape(int(len(lat_bnds)/4), 4)

    lons = []
    lats = []

    for x, y in zip(lon_bnds, lat_bnds):
        lons.append(np.mean(x))
        lats.append(np.mean(y))

    return [data, lons, lats]


class point:
    def __init__(self, x, y):
        self.x = x
        self.y = y
    def to_dict(self):
        return {'x': self.x, 'y': self.y}


# https://stackoverflow.com/questions/2049582/how-to-determine-if-a-point-is-in-a-2d-triangle
def sign(p1, p2, p3):
    return (p1.x - p3.x) * (p2.y - p3.y) - (p2.x - p3.x) * (p1.y - p3.y)

def PointInTriangle(pt, v1, v2, v3):
    d1 = sign(pt, v1, v2)
    d2 = sign(pt, v2, v3)
    d3 = sign(pt, v3, v1)

    has_neg = (d1 < 0) or (d2 < 0) or (d3 < 0)
    has_pos = (d1 > 0) or (d2 > 0) or (d3 > 0)

    return not (has_neg and has_pos)




#generate icebergs
def icb_generator(vals, ps1, ps2, ps3, icb_path, rho_ice=910, res_pism=5):
    #mu, sigma = mu_length**3*16/21/10e6, 1.2
    mu, sigma = mu_length**2*height_icb/10e6, 1.2

    points = []
    #depth=length/1.5
    height = [] #height=depth*8/7=length*8/7*2/3=length*16/21
    scaling_tmp = 1

    with tqdm(total=len(vals), file=sys.stdout, desc='create icebergs') as pbar:
        for val, p1, p2, p3 in zip(vals, ps1, ps2, ps3):
            disch_cum = val*res_pism*res_pism*10e6/rho_ice
            #N = max(int(np.absolute(disch_cum)/(10e6)/10), 1)
            N = int(np.absolute(disch_cum)/(10e6))
            if N < N_min:
                N = 0
            elif N > N_max:
                scaling_tmp = N/N_max
                N = N_max

            s = np.random.lognormal(mu, sigma, N)*10e6

            f_scale = sum(s)/np.absolute(disch_cum)
            s_ = s / f_scale / scaling_tmp
            
            length_tmp = ne.evaluate('(s_/height_icb)**(1/2)')

            #ll = np.digitize(length_tmp, slength_bins)
            ll = np.digitize(s_, svol_bins)
            for i, (s, d) in enumerate(zip(Nlength, sdepth)):
                sbin = np.where(ll==i+1)[0]
                if list(sbin):
                    if len(sbin) <= s:
                        chunk = [sbin]
                    else:
                        chunk = [sbin[k:k + s] for k in range(0, len(sbin), s)]
                    l = np.array([np.mean(s_[c]) for c in chunk])
                    s = np.array([len(c) for c in chunk])
                    if not 'length' in locals():
                        length = ne.evaluate('(l/height_icb)**(1/2)')
                        scaling = np.array([s])
                        depth = np.array([d]*len(l))
                    else:
                        length = np.append(length, ne.evaluate('(l/height_icb)**(1/2)'))
                        scaling = np.append(scaling, np.array([s]))
                        depth = np.append(depth, np.array([d]*len(l)))

                    for i in range(len(l)):
                        r1 = random.random()
                        r2 = random.random()
                        
                        lower_bound = 0.1
                        upper_bound = 0.9

                        r1 = r1 * (upper_bound - lower_bound) + lower_bound
                        r2 = r2 * (upper_bound - lower_bound) + lower_bound
                        #https://math.stackexchange.com/questions/18686/uniform-random-point-in-triangle
                        tmp_x = (1-np.sqrt(r1))*p1.x + (np.sqrt(r1)*(1-r2))*p2.x + (r2*np.sqrt(r1))*p3.x
                        tmp_y = (1-np.sqrt(r1))*p1.y + (np.sqrt(r1)*(1-r2))*p2.y + (r2*np.sqrt(r1))*p3.y
                        points.append(point(tmp_x, tmp_y))
                        #length_tmp = (np.absolute(s_[i])/height_icb)**(1/2)
                        #length.append(length_tmp)
                        ##length**2*length*16/21=s_--> length=(s_*21/16)**(1/3)
                        #length.append(np.absolute(s_[i]*21/16)**(1/3))
                        #height.append(height_icb)
                        #height.append(length[-1]*16/21)
                        #scaling.append(scaling_tmp)
            pbar.update(1)

    points_ = pd.DataFrame.from_records([p.to_dict() for p in points])
    np.savetxt(os.path.join(icb_path, "LON.dat"), points_.x.values)
    np.savetxt(os.path.join(icb_path, "LAT.dat"), points_.y.values)
    np.savetxt(os.path.join(icb_path, "LENGTH.dat"), length)
    np.savetxt(os.path.join(icb_path, "HEIGHT.dat"), depth)
    np.savetxt(os.path.join(icb_path, "SCALING.dat"), scaling)
    return [points, length]

def PointTriangle_distance(lon0, lat0, lon1, lat1, lon2, lat2, lon3, lat3):
    d1 = ne.evaluate('(lon1 - lon0)**2 + (lat1 - lat0)**2')
    d2 = ne.evaluate('(lon2 - lon0)**2 + (lat2 - lat0)**2')
    d3 = ne.evaluate('(lon3 - lon0)**2 + (lat3 - lat0)**2')
    
    dis = d1+d2+d3
    ind = np.where(dis == np.amin(dis))
   
    p1 = point(lon1[ind], lat1[ind])
    p2 = point(lon2[ind], lat2[ind])
    p3 = point(lon3[ind], lat3[ind])

    return [p1, p2, p3], ind

def find_FESOM_elem(nod2d_file, elem2d_file, lons, lats):
    points = []
    nod2d = pd.read_csv(nod2d_file, header=0, names=["lon", "lat", "coastal"], sep='\s+', index_col=0)
    elem2d = pd.read_csv(elem2d_file, header=0, names=["nod1", "nod2", "nod3"], sep='\s+')

    lon1=[] 
    lat1=[]
    lon2=[]
    lat2=[]
    lon3=[]
    lat3=[]

    lon1 = nod2d.lon[elem2d.nod1]
    lat1 = nod2d.lat[elem2d.nod1]
    lon2 = nod2d.lon[elem2d.nod2]
    lat2 = nod2d.lat[elem2d.nod2]
    lon3 = nod2d.lon[elem2d.nod3]
    lat3 = nod2d.lat[elem2d.nod3]
    
    with tqdm(total=len(lons), file=sys.stdout, desc='find FESOM elements') as pbar:
        for lon, lat in zip(lons, lats):
            tmp, ind = PointTriangle_distance(lon, lat, np.array(lon1), np.array(lat1), np.array(lon2), np.array(lat2), 
                                                    np.array(lon3), np.array(lat3))
            
            points.append(tmp[:3])
            #print("point in triangle?: ", PointInTriangle(point(lon, lat), tmp[0], tmp[1], tmp[2]))
            #print("point: [" + str(lon) + ", " + str(lat) + "]")
            #print("triangle: [", str(tmp[0].x) + ", " + str(tmp[0].y) + "], " + "[", str(tmp[1].x) + ", " + str(tmp[1].y) + "], " + "[", str(tmp[2].x) + ", " + str(tmp[2].y) + "]")
            pbar.update(1)
    return points












data, lons, lats = read_pism_file(ifile)
points = find_FESOM_elem(nod2d_file, elem2d_file, lons, lats)

ps1, ps2, ps3 = np.transpose(points)
icb_generator(data, ps1, ps2, ps3, icb_path)

#pos = find_FESOM_elem(nod2d_file, elem2d_file, xs, ys)
#
#np.savetxt("pos.csv", np.array(pos))






































def _generate_rotation_matrix(self):
    alphaEuler=50
    betaEuler=15
    gammaEuler=-90
    
    al = np.radians(alphaEuler)
    be = np.radians(betaEuler)
    ga = np.radians(gammaEuler)
   
    rotate_matrix = np.zeros((3,3))
    rotate_matrix[0,0]=np.cos(ga)*np.cos(al)-np.sin(ga)*np.cos(be)*np.sin(al)
    rotate_matrix[0,1]=np.cos(ga)*np.sin(al)+np.sin(ga)*np.cos(be)*np.cos(al)
    rotate_matrix[0,2]=np.sin(ga)*np.sin(be)
    rotate_matrix[1,0]=-np.sin(ga)*np.cos(al)-np.cos(ga)*np.cos(be)*np.sin(al)
    rotate_matrix[1,1]=-np.sin(ga)*np.sin(al)+np.cos(ga)*np.cos(be)*np.cos(al)
    rotate_matrix[1,2]=np.cos(ga)*np.sin(be)
    rotate_matrix[2,0]=np.sin(be)*np.sin(al) 
    rotate_matrix[2,1]=-np.sin(be)*np.cos(al)  
    rotate_matrix[2,2]=np.cos(be)
    
    self.rotate_matrix = rotate_matrix

def _g2r(self):
    lon1 = np.radians(self.box[0])
    lon2 = np.radians(self.box[1])
    lat1 = np.radians(self.box[2])
    lat2 = np.radians(self.box[3])

    v1_ = np.zeros((3,1))
    v1_[0]=np.cos(lat1)*np.cos(lon1)
    v1_[1]=np.cos(lat1)*np.sin(lon1)
    v1_[2]=np.sin(lat1) 
    vr1 = np.dot(self.rotate_matrix, v1_)

    v2_ = np.zeros((3,1))
    v2_[0]=np.cos(lat2)*np.cos(lon2)
    v2_[1]=np.cos(lat2)*np.sin(lon2)
    v2_[2]=np.sin(lat2) 
    vr2 = np.dot(self.rotate_matrix, v2_)
    
    self.box[0] = np.degrees(math.atan2(vr1[1], vr1[0]))
    self.box[2] = np.degrees(math.asin(vr1[2]))
    self.box[1] = np.degrees(math.atan2(vr2[1], vr2[0]))
    self.box[3] = np.degrees(math.asin(vr2[2]))

def _r2g(self, lon_r, lat_r):

    A = inv(self.rotate_matrix)

    lon_ = np.radians(lon_r)
    lat_ = np.radians(lat_r)

    v_ = np.zeros((3,1))
    v_[0]=np.cos(lat_)*np.cos(lon_)
    v_[1]=np.cos(lat_)*np.sin(lon_)
    v_[2]=np.sin(lat_) 
    vr = np.dot(A, v_)

    lon_g = np.degrees(math.atan2(vr[1], vr[0]))
    lat_g = np.degrees(math.asin(vr[2]))
    return  lon_g, lat_g

def _back_rotate(self):
    r2g_v = np.vectorize(self._r2g)

    with open(self.nod_file, 'r') as csvfile:
        nodes = pd.read_csv(csvfile, header=None, sep=r'\s* \s*', skiprows=1, engine='python')
        nodes.drop(nodes.index[0], inplace=True)
        nodes.columns = ['index', 'lon', 'lat', 'mask']
        [lon_tmp, lat_tmp] = r2g_v(np.array(nodes['lon'].values[:], dtype=float).transpose(), 
                                   np.array(nodes['lat'].values[:], dtype=float).transpose())
        nodes['lon'] = lon_tmp
        nodes['lat'] = lat_tmp
        nodes.to_csv('nodes_back_rotated.csv', sep=' ', header=True, index=False)
        self.nod_file = 'nodes_back_rotated.csv'



import time

timing_info = []


def timing(f):
    def wrap(*args):
        time1 = time.time()
        ret = f(*args)
        time2 = time.time()
        timing_info.append(
            "{:s} function took {:.3f} ms".format(
                f.__qualname__, (time2 - time1) * 1000.0
            )
        )
        return ret

    return wrap



"""
Contains functions for dealing with PBS-based batch systems
"""
import os
import subprocess
import sys
import esm_parser


class Pbs:
    """
    Deals with PBS, allowing you to check if a job is submitted, get the current job
    ID, generate the job launcher commands (e.g. ``aprun``), get the current job
    state, and check if a job is still running. Note: there is not an equivalent file
    for the slurm hostfile, because all the job information is contained in the
    launcher command.


    Parameters
    ----------
    config : dict
        Configuration dictionary containing information about the experiment and
        experiment directory.
    """

    def __init__(self, config):
        # No hostfile for PBS
        self.filename = ""
        self.path = ""

    @staticmethod
    def check_if_submitted():
        """
        Determines if a job is submitted in the currently running shell by checking for
        ``PBS_JOBID`` variable in the environment.

        Returns
        -------
        bool
            Boolean standing for submission state.
        """
        return "PBS_JOBID" in os.environ

    @staticmethod
    def get_jobid():
        """
        Gets the current ``PBS_JOBID``.

        Returns
        -------
        str or None
            If ``PBS_JOBID`` is defined returns the jobid as a string.
        """
        return os.environ.get("PBS_JOBID")

    def prepare_launcher(self, config, cluster):
        config["general"]["batch"].het_par_launcher_lines(config, cluster)

    @staticmethod
    def add_pre_launcher_lines(config, cluster, runfile):
        """
        Adds pre-launcher lines to the ``runfile``.

        Parameters
        ----------
        config : dict
            Configuration dictionary containing information about the experiment and
            experiment directory.
        runfile : io.TextIOWrapper
            File wrapper object for writing of the lines
            (``runfile.write("<your_line_here>")``).
        """
        # This changes the name of the output stream to include the $PBS_JOBID. This
        # cannot be done in the header because PBS does not support its own variables
        # to be used there (at least in the ALEPH's version).
        thisrun_logfile = config["computer"]["thisrun_logfile"].replace(
            "@jobtype@", cluster
        )
        runfile.write(f'qalter $PBS_JOBID -o {thisrun_logfile}\n')

    # TODO: remove it once it's not needed anymore (substituted by packjob)
    @staticmethod
    def write_het_par_wrappers(config):
        if config["general"].get("verbose", False):
            print(
                f"Skipping the het-par wrapper as it is not needed for {config['computer']['batch_system']}"
            )
        return config

    @staticmethod
    def het_par_headers(config, cluster, all_values):
        """
        Heterogeneous parallelization headings are not needed for PBS.
        """
        return all_values

    @staticmethod
    def get_job_state(jobid):
        """
        Returns the jobstate. See ``man qstat`` for more details.

        Parameters
        ----------
        jobid : str
            The PBS job id as displayed in, e.g. ``qstat``.

        Returns
        -------
        str or None
            The short job state.
        """
        state_command = f"qstat {str(jobid)}"

        qstat_output = subprocess.Popen(
            state_command.split(),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ).communicate()[0]
        qstat_split = str(qstat_output).split()
        if len(qstat_split) > 2:
            return qstat_split[-3]

    @staticmethod
    def job_is_still_running(jobid):
        """
        Returns ``True`` if the job is still running or queueing.

        Parameters
        ----------
        jobid : str
            The PBS job id as displayed in, e.g. ``qstat``.

        Returns
        -------
        bool
            Boolean standing for a job running or queueing.
        """
        return bool(Pbs.get_job_state(jobid))



def prepare_environment(config):
    environment_dict = {
            "TEST_IN_ENV": "testvar_in_env_exported_couple_out_vilma",
            "VILMA_TO_ICE": 1,
            "VILMA_GRID_input": config["vilma"]["grid_input"],
            "COUPLE_DIR": config["general"]["experiment_couple_dir"],
            "RUN_NUMBER_vilma": config["general"]["run_number"],
            "DATA_DIR_vilma": config["vilma"]["experiment_outdata_dir"]+"/out/",
            "RUN_DATE_STAMP": config["general"]["run_datestamp"] ,           
            }
    print(environment_dict)
    return environment_dict









"""Console script for esm_utilities."""
import sys
import click

from . import utils


@click.group()
def main(args=None):
    """Console script for esm_utilities."""
    click.echo(
        "Replace this message by putting your code into " "esm_utilities.cli.main"
    )
    click.echo("See click documentation at https://click.palletsprojects.com/")
    return 0


@main.command()
@click.argument("f")
def logfile_stats(f):
    utils.logfile_stats(f)


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



class _Getch:
    """Gets a single character from standard input.  Does not echo to the
    screen."""

    def __init__(self):
        try:
            self.impl = _GetchWindows()
        except ImportError:
            self.impl = _GetchUnix()

    def __call__(self):
        return self.impl()


def get_one_of(testlist):
    while True:
        char = getch()
        if char in testlist:
            return char


class _GetchUnix:
    def __init__(self):
        import tty, sys

    def __call__(self):
        import sys, tty, termios

        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(sys.stdin.fileno())
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch


class _GetchWindows:
    def __init__(self):
        import msvcrt

    def __call__(self):
        import msvcrt

        return msvcrt.getch()


getch = _Getch()



import argparse
import os


class Info(dict):
    """
    Dictionary subclass that contains all the information about the tests. The
    information is stored under the following `keys`:

    - ``actually_compile``: ``True`` if the test should compile or ``False`` if it
      should run a check instead.

    - ``actually_run``: ``True`` if the test should run or ``False`` if it should
      run a check instead.

    - ``bulletpoints``: when printing results or the state should it use bulletpoints
      for the items or no.

    - ``system_exit_on_errors``: triggers a system exit on errors or file differences
      so that GitHub actions can catch that as a failing test (defined by the ``-e``
      argument).

    - ``hold``: should it wait for user input after each test or not.

    - ``ignore``: lines to be ignored during comparison, loaded from
      ``<PATH>/esm_tools/src/esm_tests/resources/ignore_compare.yaml``.

    - ``ignore_user_info``: user-specific Info to remove from the files ``last_tested``
      files.

    - ``in_github``: indicates whether it is running in GitHub or not (defined by the
      ``-g`` argument).

    - ``last_tested_dir``: directory of the configuration files for the last tested
      experiments (i.e. ``<PATH>/esm_tools/src/esm_tests/resources/last_tested/``).

    - ``rm_user_info``: User-specific Info to remove from the files ``last_tested``
      files.

    - ``repo_update``: does it need to update the ``resources`` or not
      (specified by the user).

    - ``scripts``: dictionary including all information of every runscript to be run
      in the current computer.

    - ``script_dir``: directory of the ``esm_tests`` code, useful to call the
      ``resources`` and the ``resources/runscripts``.

    - ``this_computer``: name of the current computer

    - ``user``: ``esm_tests`` configurations specific to the user, as read from
      ``<PATH>/esm_tools/src/esm_tests/user_config.yaml``.

    """

    def __init__(self):
        self["script_dir"] = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "."
        )

    def argparse(self):
        """
        Parses the arguments from the ``esm_tests`` call.

        Returns
        -------
         Returns a tuple with the following entries:
        save_flag : str
            It's value can be ``true``, ``True``, ``false``, ``False``, or
            ``Not defined``.
        print_state : bool
            Indicates if ``esm_tests`` needs to report the state of the last actual
            or run tests (if ``False``).
        delete_tests : bool
            If ``True`` deletes the comp and experiment files before running the tests.
        """

        parser = argparse.ArgumentParser(
            description="Automatic testing for ESM-Tools devs"
        )
        parser.add_argument(
            "-n",
            "--no-user",
            default=False,
            help="Avoid loading user config",
            action="store_true",
        )
        parser.add_argument(
            "-c",
            "--check",
            default=False,
            help="Check mode on (does not compile or run, but produces some files that can "
            + "be compared to previous existing files in 'last_tested' folder)",
            action="store_true",
        )
        parser.add_argument(
            "-u",
            "--update",
            default=False,
            help="Updates the resources with the release branch, including runscripts"
            + "and last_tested files",
            action="store_true",
        )
        parser.add_argument(
            "-d",
            "--delete",
            default=False,
            help="Delete previous tests",
            action="store_true",
        )
        parser.add_argument(
            "-s",
            "--save",
            default="Not defined",
            help="Save files for comparisson in 'last_tested' folder",
        )
        parser.add_argument(
            "-t",
            "--state",
            default=False,
            help="Print the state stored in state.yaml",
            action="store_true",
        )
        parser.add_argument(
            "-o",
            "--hold",
            default=False,
            help="Hold before operation, to give time to check the output",
            action="store_true",
        )
        parser.add_argument(
            "-b",
            "--bulletpoints",
            default=False,
            help="bullet points for printing the results",
            action="store_true",
        )
        parser.add_argument(
            "-g",
            "--github",
            default=False,
            help="use this flag when running in GitHub servers",
            action="store_true",
        )
        parser.add_argument(
            "-e",
            "--system-exit-on-errors",
            default=False,
            help="trigger a system exit on errors or file differences so that GitHub actions can "
            + "catch that as a failing test",
            action="store_true",
        )

        args = vars(parser.parse_args())

        save_flag = args["save"]
        print_state = args["state"]
        delete_tests = args["delete"]

        self["ignore_user_info"] = args["no_user"]
        self["actually_compile"] = not args["check"]
        self["actually_run"] = not args["check"]
        self["hold"] = args["hold"]
        self["bulletpoints"] = args["bulletpoints"]
        self["repo_update"] = args["update"]
        self["in_github"] = args["github"]
        self["system_exit_on_errors"] = args["system_exit_on_errors"]

        if self["in_github"]:
            self["group_output"] = {"startg": "::group::{0}", "endg": "::endgroup::"}
        else:
            self["group_output"] = {"startg": "{0}", "endg": ""}

        return save_flag, print_state, delete_tests



import filecmp
import os
import sys
import re
import time
import pathlib

import psutil
import shutil

from . import coupler, database_actions, helpers
from .filelists import copy_files, resolve_symlinks
from . import logfiles


def run_job(config):
    config["general"]["relevant_filetypes"] = [
        "log",
        "mon",
        "outdata",
        "restart_out",
        # "bin",
        # "config",
        # "forcing",
        # "input",
        # "restart_in",
        "ignore",
        "unknown",
    ]
    helpers.evaluate(config, "tidy", "tidy_recipe")
    return config


def copy_stuff_back_from_work(config):
    config = copy_files(
        config, config["general"]["relevant_filetypes"], "work", "thisrun"
    )
    return config


def tidy_coupler(config):
    if config["general"]["standalone"] == False:
        config["general"]["coupler"].tidy(config)
    return config


def clean_run_dir(config):
    """
    This plugin allows you to clean up the ``run_${DATE}`` folders.
    To do that you can use the following variables under the
    ``general`` section of your runscript (documentation follows order
    of code as it is executed):

    * ``clean_runs``: **This is the most important variable for most
      users**. It can take the following values:
        * ``True``: removes the ``run_`` directory after each run
          (**overrides every other** ``clean_`` **option**).

        * ``False``: does not remove any ``run_`` directory (default)
          if no ``clean_`` variable is defined.

        * ``<int>``: giving an integer as a value results in deleting
          the ``run_`` folders except for the last <int> runs
          (recommended option as it allows for debugging of crashed
          simulations).

      .. Note::
         ``clean_runs: (bool)`` is incompatible with
         ``clean_this_rundir`` and ``clean_runs: (int)`` is incompatible
         with ``clean_old_rundirs_except`` (an error will be raised
         after the end of the first simulation). The functionality of
         ``clean_runs`` variable **alone will suffice most of the
         standard user requirements**. If finer tunning for the removal
         of ``run_`` directories is required you can used the following
         variables instead of ``clean_runs``.

    * ``clean_this_rundir``: (bool) Removes the entire run directory
      (equivalent to ``clean_runs: (bool)``). ``clean_this_rundir: True``
      **overrides every other** ``clean_`` **option**.

    * ``clean_old_rundirs_except``: (int) Removes the entire run
      directory except for the last <x> runs (equivalent to
      ``clean_runs: (int)``).

    * ``clean_old_rundirs_keep_every``: (int) Removes the entire
      run directory except every <x>th run. Compatible with
      ``clean_old_rundirs_except`` or ``clean_runs: (int)``.

    * ``clean_<filetype>_dir``: (bool) Erases the run directory
      for a specific filetype. Compatible with all the other options.

    * ``clean_size``: (int or float) Erases all files with size
      greater than ``clean_size``, must be specified in bytes! Compatible
      with all the other options.

    Example
    -------

    To delete all the ``run_`` directories in your experiment include this
    into your runscript:

    .. code-block:: yaml

       general:
               clean_runs: True

    To keep the last 2 ``run_`` directories:

    .. code-block:: yaml

       general:
               clean_runs: 2

    To keep the last 2 runs and every 5 runs:

    .. code-block:: yaml

       general:
               clean_old_rundirs_except: 2
               clean_old_rundirs_keep_every: 5
    """
    _clean_run_determine_user_choice(config)
    _clean_this_rundir(config)
    _clean_old_rundirs_except(config)
    _clean_old_runs_filetypes(config)
    _clean_old_runs_size(config)
    return config


def _clean_run_determine_user_choice(config):
    """
    Determine user choice from a simple switch.
    The user sets::
    general:
        clean_runs: <x>
    where ``x`` can be one of:
    * ``True`` Removes the current run dir
    * ``False`` Keeps run dir
    * ``int`` (must be >= 0) keep last ``x`` run dirs
    """
    user_clean = config["general"].get("clean_runs")
    # TODO(PG): It might be nice if these sorts of checks happened earlier
    # in the job, before it even gets to this function
    if user_clean is None:
        return  # Skip the rest of the function
    if isinstance(user_clean, bool):
        if "clean_this_rundir" not in config["general"]:
            config["general"]["clean_this_rundir"] = user_clean
        else:
            print("------------------------------------------")
            print("You have set both in your config:")
            print()
            print("general:")
            print("    clean_this_rundir: ", config["general"]["clean_this_rundir"])
            print("    clean_runs: ", user_clean)
            print()
            print("Please only use one of these!")
            print("------------------------------------------")
            sys.exit(1)
    elif isinstance(user_clean, int):
        if "clean_old_rundirs_except" not in config["general"]:
            config["general"]["clean_old_rundirs_except"] = user_clean
        else:
            print("------------------------------------------")
            print("You have set both in your config:")
            print()
            print("general:")
            print(
                "    clean_old_rundirs_except: ",
                config["general"]["clean_old_rundirs_except"],
            )
            print("    clean_runs: ", user_clean)
            print()
            print("Please only use one of these!")
            print("------------------------------------------")
            sys.exit(1)
    else:
        print("------------------------------------------")
        print("Type Error!")
        print("You have set this in your config:")
        print("general:")
        print("    clean_runs: ", user_clean)
        print()
        print("This is of type: ", type(user_clean))
        print("However, only the following types are valid:")
        print("   * boolean")
        print("   * integer (greater or equal to 0!)")
        print("Please correct that")
        print("------------------------------------------")
        sys.exit(1)


def _clean_this_rundir(config):
    if config["general"].get("clean_this_rundir", False):
        rm_r(config["general"]["thisrun_dir"])


def _clean_old_rundirs_except(config):
    all_run_folders_in_experiment = RunFolders(config)

    number_rundirs_keep_every = config["general"].get("clean_old_rundirs_keep_every")
    runs_to_keep_via_keepevery = []
    if number_rundirs_keep_every:
        try:
            assert isinstance(number_rundirs_keep_every, int)
            assert number_rundirs_keep_every >= 1
        except AssertionError:
            print("Please ensure that you use an integer in your configuration:")
            print("-------------------------------------------------------------")
            print()
            print("general:")
            print("   clean_old_rundirs_keep_every: <x>")
            print()
            print("-------------------------------------------------------------")
            print("<x> **MUST** be an integer greater or equal than 1!")
            sys.exit(1)
        runs_to_keep_via_keepevery = all_run_folders_in_experiment[
            ::number_rundirs_keep_every
        ]

    number_rundirs_to_keep = config["general"].get("clean_old_rundirs_except")
    runs_to_keep_via_end_select = []
    if number_rundirs_to_keep:
        try:
            assert isinstance(number_rundirs_to_keep, int)
            assert number_rundirs_to_keep > 1
        except AssertionError:
            print("Please ensure that you use an integer in your configuration:")
            print("-------------------------------------------------------------")
            print()
            print("general:")
            print("   clean_old_rundirs_except: <x>")
            print()
            print("-------------------------------------------------------------")
            print("<x> **MUST** be an integer greater than 1!")
            sys.exit(1)
        runs_to_keep_via_end_select = all_run_folders_in_experiment[
            -number_rundirs_to_keep:
        ]
    if number_rundirs_keep_every or number_rundirs_to_keep:
        runs_to_keep = set(runs_to_keep_via_keepevery + runs_to_keep_via_end_select)
    else:
        runs_to_keep = set(all_run_folders_in_experiment)
    runs_to_clean = set(all_run_folders_in_experiment) - runs_to_keep
    for run in list(runs_to_clean):
        rm_r(run)


def _clean_old_runs_filetypes(config):
    all_filetypes = config["general"]["all_filetypes"]
    for filetype in all_filetypes:
        if config["general"].get("clean_" + filetype + "_dir", False):
            rm_r(config["general"]["thisrun_" + filetype + "_dir"])


def _clean_old_runs_size(config):
    rmsize = config["general"].get("clean_size", False)
    if rmsize:
        flist = []
        for root, _, files in os.walk(config["general"]["thisrun_dir"]):
            for file_ in files:
                size = os.path.getsize(root + "/" + file_)
                if size >= rmsize:
                    flist.append(root + "/" + file_)
        for file_ in flist:
            os.remove(file_)


def throw_away_some_infiles(config):
    if config["general"]["run_number"] == 1:
        return config
    monitor_file = logfiles.logfile_handle
    monitor_file.write("throwing away restart_in files \n")
    for model in config["general"]["valid_model_names"]:
        print(f"{model}")
        if "thisrun_restart_in_dir" in config[model]:
            if os.path.isdir(config[model]["thisrun_restart_in_dir"]):
                for root, dirs, files in os.walk(
                    config[model]["thisrun_restart_in_dir"]
                ):
                    for name in files:
                        source = os.path.join(root, name)
                        os.remove(source)
                        print(f"Removing {source}")
    return config


def copy_all_results_to_exp(config):
    monitor_file = logfiles.logfile_handle
    monitor_file.write("Copying stuff to main experiment folder \n")
    for root, dirs, files in os.walk(config["general"]["thisrun_dir"], topdown=False):
        if config["general"]["verbose"]:
            print("Working on folder: " + root)
        if root.startswith(config["general"]["thisrun_work_dir"]) or root.endswith(
            "/work"
        ):
            if config["general"]["verbose"]:
                print("Skipping files in work.")
            continue

        for name in files:
            source = os.path.join(root, name)

            if not os.stat(source).st_size > 0:  # skip empty files
                continue

            if config["general"]["verbose"]:
                print("File: " + source)
            destination = source.replace(
                config["general"]["thisrun_dir"], config["general"]["experiment_dir"]
            )
            destination_path = destination.rsplit("/", 1)[0]
            if not os.path.exists(destination_path):
                os.makedirs(destination_path)
            if not os.path.islink(source):
                if os.path.isfile(destination):
                    if filecmp.cmp(source, destination):
                        if config["general"]["verbose"]:
                            print("File " + source + " has not changed, skipping.")
                        continue
                    else:
                        if os.path.isfile(
                            destination + "_" + config["general"]["run_datestamp"]
                        ):
                            print(
                                "Don't know where to move "
                                + destination
                                + ", file exists"
                            )
                            continue
                        else:
                            if os.path.islink(destination):
                                os.remove(destination)
                            else:
                                os.rename(
                                    destination,
                                    destination
                                    + "_"
                                    + config["general"]["last_run_datestamp"],
                                )
                            newdestination = (
                                destination + "_" + config["general"]["run_datestamp"]
                            )
                            if config["general"]["verbose"]:
                                print("Moving file " + source + " to " + newdestination)
                            os.rename(source, newdestination)
                            os.symlink(newdestination, destination)
                            continue
                try:
                    if config["general"]["verbose"]:
                        print("Moving file " + source + " to " + destination)
                    try:
                        os.rename(source, destination)
                    except:  # Fill is still open... create a hard (!) link instead
                        os.link(source, destination)

                except:
                    print(
                        ">>>>>>>>>  Something went wrong moving "
                        + source
                        + " to "
                        + destination
                    )
            else:
                linkdest = resolve_symlinks(source)
                # newlinkdest = (
                #    destination.rsplit("/", 1)[0] + "/" + linkdest.rsplit("/", 1)[-1]
                # )
                if os.path.islink(destination):
                    destdest = resolve_symlinks(source)
                    if linkdest == destdest:
                        # both links are identical, skip
                        continue
                    # os.remove(destination)
                if os.path.isfile(destination):
                    os.rename(
                        destination,
                        destination + "_" + config["general"]["last_run_datestamp"],
                    )
                os.symlink(linkdest, destination)
    return config


# Utility functions:
def rm_r(path):
    """
    Python equivalent of rm -r

    Parameters
    ----------
    path : str
        Path or directory to remove
    """
    if not os.path.exists(path):
        return
    if os.path.isfile(path) or os.path.islink(path):
        os.unlink(path)
    else:
        shutil.rmtree(path)


def size_bytes_to_human(num, suffix="B"):
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, "Yi", suffix)


# PG: BROKEN!!!
def size_human_to_bytes(s, suffix="B"):
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        num = float(s.replace(unit, ""))
        if abs(num) < 1024.0:
            return num
        num *= 1024.0
    return num


class RunFolders(list):
    """
    Logs the ``run_`` directories in ``<experiment_id>/log/run_folders.log``,
    updating it with new folders. The resulting object is a list of ``run_`` paths
    that exist or existed during the run time (even if they got deleted). This
    is useful for indexing operations such as ``<object_name>[::<interval>]``
    used when removing ``run_`` folders.
    Notes
    -----
    It keeps the folder names sorted so there is no need of sorting out of the
    object, and it also prevents the existence of duplicates.
    """

    def __init__(self, config):
        """
        The initialization of the object:
        * Loads the existing paths of the ``run_`` folders
        * Loads previous ``run_`` folder names from the logging file
        * Adds the current folder names to the logging file
        * Returns a list of ``pathlib.Path`` folder paths
        """

        # Load paths from ``config``
        self.exp_dir = config["general"]["experiment_dir"]
        self.log_path = self.exp_dir + "/log/run_folders.log"

        # Load existing folders
        self.current_folders = [
            folder for folder in os.listdir(self.exp_dir) if folder.startswith("run_")
        ]
        self.current_folders = [
            self.exp_dir + "/" + folder for folder in self.current_folders
        ]

        # Check if the ``run_folders.log`` file exists, and if not, create it
        if not os.path.exists(self.log_path):
            with open(self.log_path, "w") as log_file:
                pass

        # Load previous run names from ``run_folders.log``
        self.folders = []
        self.load()

        # Add current folders
        self.update()

        # Add folders to the list
        for folder in self.folders:
            super().append(pathlib.Path(folder))

    def load(self):
        """
        Loads the existing paths of the ``run_`` folders.
        """
        with open(self.log_path, "r") as log_file:
            for folder in log_file.readlines():
                self.folders.append(folder.strip())

    def save(self):
        """
        Saves all folder names.
        """
        with open(self.log_path, "w") as log_file:
            log_file.writelines([folder + "\n" for folder in self.folders])

    def update(self):
        """
        Updates the folders read from the log file with the currently existing
        folders, removes duplicates, sorts them and save them into the log file.
        """
        # Update with ``self.curren_folders``
        for folder in self.current_folders:
            self.folders.append(folder)
        # Remove duplicates
        self.folders = list(dict.fromkeys(self.folders))
        # Sort folders
        self.folders.sort()
        # Save to the log file
        self.save()



#!/usr/bin/env python3
# NOTE(PG): I hate how much duplication there is in this....

import os
import sys
import pkg_resources

import yaml


def _dist_is_editable(dist):
    """Is distribution an editable install?"""
    for path_item in sys.path:
        egg_link = os.path.join(path_item, dist.replace("_", "-") + ".egg-link")
        if os.path.isfile(egg_link):
            return True
    return False


EDITABLE_INSTALL = _dist_is_editable("esm_tests")
"""
bool : Shows if the installation is installed in editable mode or not.
"""


# State YAML:
def get_state_yaml():
    if EDITABLE_INSTALL:
        return yaml.safe_load(pkg_resources.resource_string("resources", "state.yaml"))
    return yaml.safe_load(
        pkg_resources.resource_string("esm_tests.resources", "state.yaml")
    )


# State YAML path:
def get_state_yaml_path():
    if EDITABLE_INSTALL:
        return pkg_resources.resource_filename("resources", f"state.yaml")
    return yaml.safe_load(
        pkg_resources.resource_filename("esm_tests.resources", "state.yaml")
    )


# Ignore Compare YAML:
def get_ignore_compare_yaml():
    if EDITABLE_INSTALL:
        return yaml.safe_load(
            pkg_resources.resource_string("resources", "ignore_compare.yaml")
        )
    return yaml.safe_load(
        pkg_resources.resource_string("esm_tests.resources", "ignore_compare.yaml")
    )


# Last Tested:
def get_last_tested(f):
    if EDITABLE_INSTALL:
        return pkg_resources.resource_string("resources", f"last_tested/{f}")
    return pkg_resources.resource_string("esm_tests.resources", f"last_tested/{f}")


# Last Tested dir:
def get_last_tested_dir():
    if EDITABLE_INSTALL:
        return pkg_resources.resource_filename("resources", f"last_tested/")
    return pkg_resources.resource_filename("esm_tests.resources", f"last_tested/")


# Runscripts:
def get_runscripts(f):
    if EDITABLE_INSTALL:
        return pkg_resources.resource_string("resources", f"runscripts/{f}")
    return pkg_resources.resource_string("esm_tests.resources", f"runscripts/{f}")


# Runscripts dir:
def get_runscripts_dir():
    if EDITABLE_INSTALL:
        return pkg_resources.resource_filename("resources", f"runscripts/")
    return pkg_resources.resource_filename("esm_tests.resources", f"runscripts/")



import os
import questionary

from loguru import logger

from .test_utilities import sh
from esm_parser import user_error


def update_resources_submodule(info, verbose=True):
    """
    Initiates and updates the module ``esm_tests_info``.

    Parameters
    ----------
    info : esm_tests.Info
        Dictionary with the general info about the tests.
    verbose : bool
        Verbose option.
    """

    check_resources(info, verbose)

    if info["repo_update"]:
        resources_branch = "release"
        text = f"Updating the resources repo with '{resources_branch}' branch"

        logger.info(text)
        logger.info("-" * len(text))

        # Initiate the subodule
        os.chdir(info["script_dir"])
        sh("git submodule init", verbose=verbose)
        sh("git submodule update", verbose=verbose)
        os.chdir(f"{info['script_dir']}/resources")

        # Pull the desired branch of resources
        try:
            sh(f"git checkout {resources_branch}")
            sh("git pull")
        except:
            logger.error(f"Pull of {resources_branch} branch not possible")
            sys.exit(1)


def check_resources(info, verbose=True):
    """
    Check if the submodule ``resources`` (``esm_tests_info``) is installed yet, and if
    not, asks the user whether it needs to install it or not. If the user wants to
    install it sets ``info["repo_update"] = True``.

    Parameters
    ----------
    info : esm_tests.Info
        Dictionary with the general info about the tests.
    verbose : bool
        Verbose option.

    Notes
    -----
    Missing resources for ESM-Tests error : esm_parser.user_error
        Returns an ``esm_parser.user_error`` if the user decides not to install the
        submodule.
    """
    resources_folder = f"{info['script_dir']}/resources/"
    dir_resources = os.listdir(resources_folder)
    if not (
        "last_tested" in dir_resources
        and "runscripts" in dir_resources
        and not info.get("repo_update", False)
    ):
        install_resources = questionary.confirm(
            "The resources submodule for esm_tests is not installed yet (i.e. the "
            f"{resources_folder} folder does not include the last_tested or the "
            "runscripts folder). Would you like to install the submodule now?",
        ).ask()
        if install_resources:
            info["repo_update"] = True
        else:
            user_error(
                "Missing resources for ESM-Tests",
                "ESM-Tests needs the esm_tests_info submodule. Please, run "
                "``esm_tests -u`` or accept the installation of the submodule in the "
                "previous questionary.",
            )
        print()


def info_repo():
    """
    Return a git object for the esm_tests_info repo
    """
    pass


def esm_tools_repo():
    """
    Return a git object for the esm_tools repo
    """
    pass


def checkout_info():
    """
    Checks out the release branch in the resources repo and pulls
    """
    pass


def commit_info_changes():
    """
    commits
    """



"""Top-level package for ESM Tests."""

__author__ = """Miguel Andres-Martinez"""
__email__ = "miguel.andres-martinez@awi.de"
__version__ = "6.17.5"

from .initialization import *
from .read_shipped_data import *
from .test_utilities import *
from .tests import *



import sys


class oasis:
    def __init__(
        self,
        nb_of_couplings=1,
        coupled_execs=["echam", "fesom"],
        runtime=1,
        debug_level=1,
        nnorest="F",
        mct_version="4.0",
        lucia=False,
    ):
        if isinstance(mct_version, tuple):
            pass
        elif isinstance(mct_version, int):
            mct_version = (mct_version, 0)
        elif isinstance(mct_version, float):
            mct_version = tuple(int(x) for x in str(mct_version).split("."))
        # PG: Fixup version to be tuples:
        elif isinstance(mct_version, str):
            mct_version = tuple(int(x) for x in mct_version.split("."))
        else:
            print(
                "Init of Oasis needs the argument mct_version to be either a tuple or a string!",
                flush=True,
            )
            sys.exit(1)
        self.namcouple = [
            "# This namcouple was automatically generated by the esm-tools (Python)"
        ]
        self.namcouple += [" $NFIELDS", "            " + str(nb_of_couplings), " $END"]
        exec_entry = ""
        for exe in coupled_execs:
            exec_entry = exec_entry + " " + exe
        exec_entry = str(len(coupled_execs)) + exec_entry
        self.namcouple += [" $NBMODEL", "            " + str(exec_entry), " $END"]
        self.namcouple += [" $RUNTIME", "           " + str(runtime), " $END"]
        # seb-wahl: add lucia support
        if lucia:
            self.namcouple += [" $NLOGPRT", "           " + "1 -1", " $END"]
        else:
            self.namcouple += [" $NLOGPRT", "           " + str(debug_level), " $END"]
        if mct_version >= (4, 0):
            # If true, OASIS can start without restart files
            self.namcouple += [" $NNOREST", "           " + str(nnorest), " $END "]
        self.namcouple += [" $STRINGS"]
        self.namcouple += [
            "###############################################################################"
        ]
        self.namcouple += [
            "###############################################################################"
        ]
        self.next_coupling = 1
        self.name = "oasis3mct"

    def add_input_coupling(self, field_name, freq, field_filepath):
        self.namcouple += ["#"]
        nb = self.next_coupling
        self.namcouple += [
            f"{field_name} {field_name} {nb} {freq} 0 {field_filepath} INPUT"
        ]
        self.namcouple += "#"
        self.next_coupling += 1

    def add_coupling(
        self,
        lefts,
        lgrid,
        rights,
        rgrid,
        direction,
        transformation,
        restart_file,
        time_step,
        lresume,
        export_mode="DEFAULT",
    ):
        import sys

        self.namcouple += ["#"]

        nb = self.next_coupling

        left = sep = ""
        for lefty in lefts:

            restart_out_file = lefty + "_"

            left += sep + lefty
            self.next_coupling += 1
            sep = ":"

        right = sep = ""
        for righty in rights:
            right += sep + righty
            sep = ":"

        if export_mode == "DEFAULT":
            if lresume == False:
                export_mode = "EXPOUT"
            else:
                export_mode = "EXPORTED"

        if lresume == False:
            lag = str(0)
        else:
            lag = direction.get("lag", "0")

        # if a transformation method for CONSERV (e.g. GLOBAL) is set below,
        # increase seq (=number of lines describing the transformation) by 1
        # this part was broke, confused seq with nb of trafo lines

        seq = int(direction.get("seq", "2"))
        # if transformation.get("postprocessing", {}).get("conserv", {}).get("method"):
        #    seq += 1

        p_rgrid = p_lgrid = "0"
        if "number_of_overlapping_points" in rgrid:
            p_rgrid = str(rgrid["number_of_overlapping_points"])
        if "number_of_overlapping_points" in lgrid:
            p_lgrid = str(lgrid["number_of_overlapping_points"])

        trafo_line = ""
        trafo_details = []

        alltimes = transformation.get("time_transformation", "bla")
        if not type(alltimes) == list:
            alltimes = [alltimes]
        for time in alltimes:
            detail_line = ""
            if time.lower() in ["instant", "accumul", "average", "t_min", "t_max"]:
                trafo_line = "LOCTRANS"
                detail_line = time.upper()
                trafo_details.append(detail_line.strip())

        allpres = transformation.get("preprocessing", "bla")
        if not type(allpres) == list:
            allpres = [allpres]
        for pre in allpres:
            detail_line = ""
            if type(pre) == dict:
                pre = list(pre.keys())[0]
            if pre.lower() == "checkin":
                trafo_line += " CHECKIN"
                detail_line = "INT = 1"
                trafo_details.append(detail_line.strip())
            elif pre.lower() == "blasold":
                trafo_line += " BLASOLD"
                coefficient = transformation["preprocessing"][pre].get("xmult", None)
                if not coefficient:
                    print(
                        "xmult needs to be defined for preprocessing BLASOLD",
                        flush=True,
                    )
                    sys.exit(2)
                add_scalar = transformation["preprocessing"][pre].get(
                    "add_scalar", None
                )
                if not add_scalar:
                    print(
                        "add_scalar needs to be defined (0 or 1) for preprocessing BLASOLD",
                        flush=True,
                    )
                    sys.exit(2)
                detail_line = str(coefficient) + " " + str(add_scalar)
                trafo_details.append(detail_line.strip())
                if str(add_scalar) == "1":
                    scalar_to_add = transformation["preprocessing"][pre].get(
                        "scalar_to_add", None
                    )
                    if not add_scalar:
                        print(
                            "scalar_to_add needs to be defined if add_scalar is set to  1 for preprocessing BLASOLD",
                            flush=True,
                        )
                        sys.exit(2)
                    detail_line = " CONSTANT" + str(add_scalar)
                    trafo_details.append(detail_line.strip())

        alltrans = transformation.get("remapping", {"bla": "blub"})
        if not type(alltrans) == list:
            alltrans = [alltrans]
        for thistrans in alltrans:
            (trans, transform) = list(thistrans.items())[0]
            detail_line = ""
            if "mapping" == trans.lower():
                trafo_line += " MAPPING"
                mapname = transform.get("mapname", None)
                if not mapname:
                    print(
                        "mapname needs to be defined for transformation MAPPING",
                        flush=True,
                    )
                    sys.exit(2)
                maploc = transform.get("map_regrid_on", "")
                mapstrat = transform.get("mapstrategy", "")
                detail_line = mapname + " " + maploc + " " + mapstrat
                trafo_details.append(detail_line.strip())

            elif trans.lower() in [
                "distwgt",
                "bicubic",
                "bilinear",
                "gauswgt",
                "conserv",
                "loccunif",
            ]:
                trafo_line += " SCRIPR"
                srcgridtype = str(rgrid["oasis_grid_type"]).upper()
                search_bin = transform.get("search_bin", None)
                if not search_bin:
                    print(
                        "search_bin (LATITUDE or LATLON) needs to be defined for transformations DISTWGT, GAUSWGT, BILINEAR, BICUBIC, LOCCUNIF",
                        flush=True,
                    )
                    sys.exit(2)
                bins = transform.get("nb_of_search_bins", "1")
                detail_line = (
                    trans.upper()
                    + " "
                    + srcgridtype.upper()
                    + " SCALAR "
                    + search_bin.upper()
                    + " "
                    + str(bins)
                )
                if trans.lower() in ["distwgt", "gauswgt", "loccunif"]:
                    nb_of_neighbours = transform.get("nb_of_neighbours", None)
                    if not nb_of_neighbours:
                        print(
                            "nb_of_neighbours needs to be defined for transformations DISTWGT, GAUSWGT and LOCCUNIF",
                            flush=True,
                        )
                        sys.exit(2)
                    detail_line += " " + str(nb_of_neighbours)
                if trans.lower() == "gauswgt":
                    weight = transform.get("weight", None)
                    if not weight:
                        print(
                            "weight needs to be defined for transformation GAUSWGT",
                            flush=True,
                        )
                        sys.exit(2)
                    detail_line += " " + str(weight)
                if trans.lower() == "conserv":
                    normalization = transform.get("normalization", None)
                    if not normalization:
                        print(
                            "normalization (FRACAREA, DESTAREA or FRACNNEI) needs to be defined for transformations CONSERV",
                            flush=True,
                        )
                        sys.exit(2)
                    order = transform.get("order", None)
                    if not order:
                        print(
                            "order (FIRST or SECOND) needs to be defined for transformation CONSERV",
                            flush=True,
                        )
                        sys.exit(2)
                    detail_line += " " + normalization.upper() + " " + order.upper()
                trafo_details += [detail_line.strip()]

        allpost = transformation.get("postprocessing", "bla")
        if not type(allpost) == list:
            allpost = list(allpost)
        for post in allpost:
            detail_line = ""
            if post.lower() == "conserv":
                trafo_line += " CONSERV"
                method = transformation["postprocessing"][post].get("method", None)
                if not method:
                    print(
                        " a method (GLOBAL, GLBPOS, BASBAL or BASPOS) needs to be defined for postprocessing CONSERV",
                        flush=True,
                    )
                    sys.exit(2)
                algorithm = transformation["postprocessing"][post].get("algorithm", "")
                detail_line = method.upper() + " " + algorithm.lower()
                trafo_details.append(detail_line.strip())
            elif post.lower() == "checkout":
                trafo_line += " CHECKOUT"
                detail_line = "INT = 1"
                trafo_details.append(detail_line.strip())
            elif post.lower() == "blasnew":
                trafo_line += " BLASNEW"
                coefficient = transformation["postprocessing"][post].get("xmult", None)
                if not coefficient:
                    print(
                        "xmult needs to be defined for postprocessing BLASNEW",
                        flush=True,
                    )
                    sys.exit(2)
                add_scalar = transformation["postprocessing"][post].get(
                    "add_scalar", None
                )
                if not add_scalar:
                    print(
                        "add_scalar needs to be defined (0 or 1) for preprocessing BLASOLD",
                        flush=True,
                    )
                    sys.exit(2)
                detail_line = str(coefficient) + " " + str(add_scalar)
                trafo_detail.append(detail_line.strip())
                if str(add_scalar) == "1":
                    scalar_to_add = transformation["postprocessing"][post].get(
                        "scalar_to_add", None
                    )
                    if not add_scalar:
                        print(
                            "scalar_to_add needs to be defined if add_scalar is set to  1 for postprocessing BLASNEW",
                            flush=True,
                        )
                        sys.exit(2)
                    detail_line = " CONSTANT" + str(add_scalar)
                    trafo_details.append(detail_line.strip())

        # Remove the first space in cases such as trafo_line = ' SCRIPR CONSERV'
        if trafo_line[0] == " ":
            trafo_line = trafo_line[1:]

        nb_of_trafo_lines = len(trafo_details)

        self.namcouple += [
            right
            + " "
            + left
            + " "
            + str(nb)
            + " "
            + str(time_step)
            + " "
            + str(nb_of_trafo_lines)
            + " "
            + str(restart_file)
            + " "
            + export_mode
        ]
        if lgrid and rgrid:
            self.namcouple += [
                str(rgrid["nx"])
                + " "
                + str(rgrid["ny"])
                + " "
                + str(lgrid["nx"])
                + " "
                + str(lgrid["ny"])
                + " "
                + rgrid["name"]
                + " "
                + lgrid["name"]
                + " LAG="
                + str(lag)
            ]

        self.namcouple += ["P " + p_rgrid + " P " + p_lgrid]
        self.namcouple += [trafo_line]
        for line in trafo_details:
            self.namcouple += [line]

        self.namcouple += ["#"]
        self.namcouple += ["#"]
        self.namcouple += ["#"]
        self.namcouple += [
            "###############################################################################"
        ]

    def print_config_files(self):
        for line in self.namcouple:
            print(line, flush=True)

    def add_output_file(self, lefts, rights, leftmodel, rightmodel, config):
        out_file = []

        coupling = self.next_coupling

        if self.next_coupling < 10:
            this_coupling = "0" + str(coupling)
        else:
            this_coupling = str(coupling)

        for lefty in lefts:
            out_file.append(lefty + "_" + leftmodel + "_" + this_coupling + ".nc")
        for righty in rights:
            out_file.append(righty + "_" + rightmodel + "_" + this_coupling + ".nc")

        self.next_coupling += 1

        if not "outdata_files" in config:
            config["outdata_files"] = {}
        if not "outdata_in_work" in config:
            config["outdata_in_work"] = {}
        if not "outdata_sources" in config:
            config["outdata_sources"] = {}

        for thisfile in out_file:

            config["outdata_files"][thisfile] = thisfile
            config["outdata_in_work"][thisfile] = thisfile
            config["outdata_sources"][thisfile] = thisfile

    def add_restart_files(self, restart_file, fconfig):
        config = fconfig[self.name]
        gconfig = fconfig["general"]
        # enddate = "_" + str(gconfig["end_date"].year) + str(gconfig["end_date"].month) + str(gconfig["end_date"].day)
        # parentdate = "_" + str(config["parent_date"].year) + str(config["parent_date"].month) + str(config["parent_date"].day)
        enddate = "_" + gconfig["end_date"].format(
            form=9, givenph=False, givenpm=False, givenps=False
        )
        parentdate = "_" + config["parent_date"].format(
            form=9, givenph=False, givenpm=False, givenps=False
        )

        if not "restart_out_files" in config:
            config["restart_out_files"] = {}
        if not "restart_out_in_work" in config:
            config["restart_out_in_work"] = {}
        if not "restart_out_sources" in config:
            config["restart_out_sources"] = {}

        if not "restart_in_files" in config:
            config["restart_in_files"] = {}
        if not "restart_in_in_work" in config:
            config["restart_in_in_work"] = {}
        if not "restart_in_sources" in config:
            config["restart_in_sources"] = {}

        config["restart_out_files"][restart_file] = restart_file
        config["restart_out_files"][restart_file + "_recv"] = restart_file + "_recv"

        config["restart_out_in_work"][restart_file] = restart_file  # + enddate
        config["restart_out_in_work"][restart_file + "_recv"] = (
            restart_file + "_recv"
        )  # + enddate

        config["restart_out_sources"][restart_file] = restart_file
        config["restart_out_sources"][restart_file + "_recv"] = restart_file + "_recv"

        config["restart_in_files"][restart_file] = restart_file
        config["restart_in_in_work"][restart_file] = restart_file
        if not restart_file in config["restart_in_sources"]:
            config["restart_in_sources"][restart_file] = restart_file

    def prepare_restarts(self, restart_file, all_fields, models, config):
        enddate = "_" + config["general"]["end_date"].format(
            form=9, givenph=False, givenpm=False, givenps=False
        )
        # enddate = "_" + str(config["general"]["end_date"].year) + str(config["general"]["end_date"].month) + str(config["general"]["end_date"].day)
        import glob
        import os
        import subprocess

        print("Preparing oasis restart files from initial run...", flush=True)
        # Assign an exe per model
        exes = [config[model]["executable"] for model in models]
        print(restart_file, all_fields, models, exes, flush=True)
        cwd = os.getcwd()
        os.chdir(config["general"]["thisrun_work_dir"])
        filelist = ""
        # Loop through the fields and their corresponding models and exes
        for field, model, exe in zip(all_fields, models, exes):
            print(field + "-" + model, flush=True)
            thesefiles = glob.glob(field + "_" + exe + "_*.nc")
            print(thesefiles, flush=True)
            for thisfile in thesefiles:
                print("cdo showtime " + thisfile + " 2>/dev/null | head -n 1 | wc -w", flush=True)
                lasttimestep = (
                    subprocess.check_output(
                        "cdo showtime " + thisfile + " 2>/dev/null | head -n 1 | wc -w", shell=True
                    )
                    .decode("utf-8")
                    .rstrip()
                )
                # print (lasttimestep)

                print(
                    "cdo -O seltimestep,"
                    + str(lasttimestep)
                    + " "
                    + thisfile
                    + " onlyonetimestep.nc",
                    flush=True,
                )
                os.system(
                    "cdo -O seltimestep,"
                    + str(lasttimestep)
                    + " "
                    + thisfile
                    + " onlyonetimestep.nc"
                )
                print(
                    "ncwa -O -a time onlyonetimestep.nc notimestep_" + field + ".nc",
                    flush=True,
                )
                os.system(
                    "ncwa -O -a time onlyonetimestep.nc notimestep_" + field + ".nc"
                )
                filelist += "notimestep_" + field + ".nc "
                print(filelist)
        # MA: -O flag added to overwrite oasis restart files in case oasis creats them
        # before (i.e. when using LOCTRANS)
        if os.path.isfile(restart_file) and config["general"]["verbose"]:
            print(f"{restart_file} already exits, overwriting", flush=True)
        print("cdo -O merge " + filelist + " " + restart_file, flush=True)  # + enddate)
        os.system("cdo -O merge " + filelist + " " + restart_file)  # + enddate)
        rmlist = glob.glob("notimestep*")
        rmlist.append("onlyonetimestep.nc")
        for rmfile in rmlist:
            print("rm " + rmfile, flush=True)
            os.system("rm " + rmfile)
        os.chdir(cwd)

    def finalize(self, destination_dir):
        self.namcouple += [" $END"]
        endline = ""
        with open(destination_dir + "/namcouple", "w") as namcouple:
            for line in self.namcouple:
                namcouple.write(endline)
                namcouple.write(line)
                endline = "\n"
